import functools
import itertools
from collections import defaultdict
from typing import Any, Callable, Dict, Iterable, List, Literal, Optional, Tuple, Union

from docutils.nodes import Element
from sphinx.addnodes import pending_xref
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.util.inspect import signature_from_str, stringify_signature

from aioqbt.bittorrent import InfoHash, InfoHashes, InfoHashesOrAll


@functools.cache
def stringify_annotation(anno: Any, mode: str) -> str:
    from sphinx.util.typing import stringify_annotation

    return stringify_annotation(anno, mode)


class ReplacementAnnotations:
    """
    Replacement annotation

    Typehints are stringified to compare, substitute, and cached.
    """

    def __init__(self, mode: str):
        """
        :param mode: stringify() mode
        """
        assert mode in {"smart", "fully-qualified", "fully-qualified-except-typing"}, mode
        self._mode = mode

        # param -> (qualname prefix, target, replacement)
        self._registry: Dict[str, List[Tuple[str, str, str]]] = defaultdict(list)

        # (prefix, param) -> target -> replacement
        self._table: Dict[Tuple[str, str], Dict[str, str]] = defaultdict(dict)

    @classmethod
    def _get_prefixes(cls, name: str) -> List[str]:
        if not name:
            return [""]
        parts = name.split(".")
        prefixes = list(itertools.accumulate(parts, lambda a, b: f"{a}.{b}"))
        prefixes.reverse()
        prefixes.append("")
        return prefixes

    def add_rule(self, prefix: str, param: str, target: Any, replacement: Any):
        """
        :param prefix: qualname prefix
        :param param: parameter name
        :param target: original type hint
        :param replacement: replacement type hint
        """
        target = stringify_annotation(target, self._mode)
        replacement = stringify_annotation(replacement, self._mode)

        item = (
            prefix,
            target,
            replacement,
        )
        self._registry[param].append(item)

        current = self._table[(prefix, param)].setdefault(target, replacement)
        if current != replacement:
            raise RuntimeError(f"Duplicate rule {prefix!r} {param!r} {target!r} {replacement!r}")

    def find_replacement(self, name: str, param: str, target: Any) -> Optional[str]:
        """
        Find a replacement for annotation ``target`` located
        in ``name`` and ``param`` or return ``None`` if unavailable.
        """

        for prefix in self._get_prefixes(name):
            loc = (prefix, param)

            if loc not in self._table:
                continue

            result = self._table[loc].get(stringify_annotation(target, self._mode))
            if result is None:
                continue

            return result

        return None


def _builder_inited(app: Sphinx):
    if app.config.autodoc_typehints_format == "short":
        mode = "smart"
        tilde = "~"
    else:
        mode = "fully-qualified"
        tilde = ""

    ra = ReplacementAnnotations(mode)

    qn_hash = f"{tilde}aioqbt.bittorrent.InfoHash"
    qn_hashes = Iterable[qn_hash]
    qn_hashes_all = Union[qn_hashes, Literal["all"]]

    table = [
        ("aioqbt", "hash", InfoHash, qn_hash),
        ("aioqbt", "hashes", InfoHashes, qn_hashes),
        ("aioqbt", "hashes", InfoHashesOrAll, qn_hashes_all),
        ("aioqbt", "id", InfoHashes, qn_hashes),
        ("aioqbt", "id", InfoHashesOrAll, qn_hashes_all),
    ]

    for package, param, target, repl in table:
        ra.add_rule(package, param, target, repl)
        ra.add_rule(package, param, Optional[target], Optional[repl])

    app.env._monkeypatch_repl_annos = ra


def _process_signature(
    app: Sphinx,
    what: str,
    name: str,
    obj: Callable,
    options: Dict[str, Any],
    signature: Optional[str],
    return_annotation: Optional[str],
):
    if what not in {"function", "method"}:
        return

    if signature in (None, "", "()") and return_annotation in (None, ""):
        # trivial
        return

    ra: ReplacementAnnotations = app.env._monkeypatch_repl_annos

    sig = signature_from_str(signature)
    parameters = list(sig.parameters.values())
    changed = False

    for i, param in enumerate(parameters):
        if param.annotation is None:
            continue

        repl = ra.find_replacement(name, param.name, param.annotation)
        if repl is None:
            continue

        changed = True
        parameters[i] = param.replace(annotation=repl)

    new_return_anno = return_annotation

    if return_annotation is not None:
        new_return_anno = ra.find_replacement(name, "return", return_annotation)
        if new_return_anno is None:
            new_return_anno = return_annotation
        else:
            changed = True

    if not changed:
        return

    sig = sig.replace(parameters=parameters)
    new_signature = stringify_signature(sig)

    if new_signature == signature and new_return_anno == return_annotation:
        return

    return new_signature, new_return_anno


def _missing_reference(
    app: Sphinx,
    env: BuildEnvironment,
    node: pending_xref,
    contnode: Element,
):
    # Redirect missing "class" xref to "data" in Python domain
    # which may be type hints
    if node.get("refdomain") != "py":
        return None
    reftarget = node.get("reftarget")
    if reftarget is None or not reftarget.startswith("aioqbt."):
        return None
    if node.get("reftype") not in ("class",):
        return None

    domain = env.get_domain("py")
    result = domain.resolve_xref(
        env,
        node["refdoc"],
        app.builder,
        "data",
        reftarget,
        node,
        contnode,
    )
    return result


def setup(app: Sphinx):
    app.setup_extension("sphinx.ext.autodoc")

    app.connect("builder-inited", _builder_inited)
    app.connect("autodoc-process-signature", _process_signature)
    app.connect("missing-reference", _missing_reference)

    return {
        "version": "1.0.0",
        "env_version": 1,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import pytest
from helper.client import LoginInfo

from aioqbt import exc
from aioqbt.client import create_client


@pytest.mark.asyncio
async def test_login_failed(mock_login: LoginInfo):
    with pytest.raises(exc.LoginError):
        await create_client(
            url=mock_login.url,
            username="badUsername",
            password="badPassword",
        )


@pytest.mark.asyncio
async def test_logout(mock_login: LoginInfo):
    client = await create_client(
        url=mock_login.url,
        username=mock_login.username,
        password=mock_login.password,
    )

    async with client:
        await client.auth.logout()

        with pytest.raises(exc.ForbiddenError):
            await client.app.version()



"""
"""
import asyncio
import logging
import types
from abc import ABCMeta
from ssl import SSLContext
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    NoReturn,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import aiohttp
from aiohttp.hdrs import RETRY_AFTER as _RETRY_AFTER
from typing_extensions import Self
from yarl import URL

from aioqbt import exc
from aioqbt._compat import cached_property
from aioqbt._paramdict import ParamDict
from aioqbt.mapper import ObjectMapper
from aioqbt.version import APIVersion, ClientVersion

if TYPE_CHECKING:
    from aioqbt.api import (
        AppAPI,
        AuthAPI,
        LogAPI,
        RSSAPI,
        SearchAPI,
        SyncAPI,
        TorrentsAPI,
        TransferAPI,
    )

__all__ = (
    "APIClient",
    "APIGroup",
    "create_client",
)

T = TypeVar("T")
K = TypeVar("K")


class APIClient:
    """
    Represent a remote qBittorrent client.
    """

    def __init__(
        self,
        base_url: str,
        *,
        mapper: Optional[ObjectMapper] = None,
        http: Optional[aiohttp.ClientSession] = None,
        ssl: Optional[SSLContext] = None,
        client_version: Optional[ClientVersion] = None,
        api_version: Optional[APIVersion] = None,
        logout_when_close: Optional[bool] = None,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        if logger is None:
            logger = logging.getLogger(
                "%s.%s" % (type(self).__module__, type(self).__qualname__),
            )

        base_url = base_url.rstrip("/")

        if mapper is None:
            mapper = ObjectMapper()

        if http is None:
            http = aiohttp.ClientSession()
            http_owner = True
        else:
            http_owner = False

        context: Dict[Any, Any] = {
            "mapper": mapper,
            "client_version": client_version,
            "api_version": api_version,
        }

        self._logger = logger
        self._mapper = mapper
        self._context = context
        self._http: Optional[aiohttp.ClientSession] = http
        self._http_owner = http_owner
        self._ssl = ssl
        self.base_url = base_url
        self._retry_statuses = {
            429,  # Too many requests
            503,  # Service unavailable
            502,  # Bad gateway: reverse proxy may be overloaded
        }
        self._logout_when_close = logout_when_close

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.base_url!r}>"

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[types.TracebackType],
    ) -> None:
        await self.close()

    def is_closed(self) -> bool:
        """Tell whether client is closed"""
        return self._http is None

    @property
    def closed(self) -> bool:
        # deprecated, prefer `is_closed()` to `closed` property
        return self._http is None

    async def close(self) -> None:
        """
        Close client.

        Release/detach resources acquired by client.
        """

        if self._http is None:
            return

        if self._logout_when_close:
            try:
                await self.auth.logout()
            except exc.ForbiddenError:
                pass

        # break cycle references to help GC.
        vars_dict = vars(self)

        for name in ("app", "auth", "log", "sync", "torrents", "transfer"):
            if name not in vars_dict:
                continue
            group = getattr(self, name, None)
            if isinstance(group, APIGroup):
                group._close()
                vars_dict.pop(name, None)

        # Close if owned
        if self._http_owner:
            await self._http.close()

        # Detach ClientSession
        self._http = None

    async def request(
        self,
        method: str,
        endpoint: str,
        *,
        params: Any = None,
        data: Any = None,
        max_attempts: int = 3,
        retry_delay: float = 5,
        ssl: Optional[SSLContext] = None,
        **kwargs: object,
    ) -> aiohttp.ClientResponse:
        """
        Send an HTTP request and return a response object.

        Argument ``method`` specifies the HTTP method (e.g. ``GET``)
        while ``endpoint`` the API endpoint (e.g. ``torrents/info``).

        ``params`` forms the query string of the request URL.
        ``data`` is the payload in the request body.
        See the underlying :meth:`ClientSession.request <aiohttp.ClientSession.request>`
        for their allowed values.

        ``max_attempts`` is the maximum number of attempts.
        Retry is performed if two additional conditions are satisfied:

        * ``GET`` or ``HEAD`` requets
        * Remote disconnection, or repsonse status
          429 (Too many requests), 503 (Service unavailable), or 502 (Bad gateway).

        The result is :class:`aiohttp.ClientResponse`, and should be used in ``async with``.

        :param str method: HTTP method.
        :param str endpoint: API endpoint.
        :param params: parameters in query string
        :param data: data in request body
        :param int max_attempts: maximum number of attempts
        :param float retry_delay: maximum delay between attempts
        :param ssl: :class:`~ssl.SSLContext`, optional
        :raise APIError: API errors (non-``200`` status).
        :raise aiohttp.ClientError: connection errors
        :return: :class:`~aiohttp.ClientResponse`
        """

        if max_attempts <= 0:
            raise ValueError(f"max_attempts <= 0: {max_attempts!r}")
        elif method.upper() not in ("GET", "HEAD"):
            max_attempts = 1

        if self._http is None:
            raise RuntimeError("closed client")

        url = self.base_url + "/" + endpoint.lstrip("/")

        if isinstance(params, ParamDict):
            params = params.to_dict()

        if isinstance(data, ParamDict):
            data = data.to_dict()

        if ssl is None:
            ssl = self._ssl

        attempt_count = 1
        while True:
            resp: Optional[aiohttp.ClientResponse] = None
            resp_body: bytes = b""

            try:
                resp = await self._http.request(
                    method,
                    url,
                    params=params,
                    data=data,
                    ssl=ssl,
                    raise_for_status=False,
                    **kwargs,
                )

                if resp.status == 200:
                    return resp

                # treat all status except 200 an error
                # Read the response before release the response
                resp_body = await resp.read()
                resp.release()

                raise aiohttp.ClientResponseError(
                    resp.request_info,
                    resp.history,
                    status=resp.status,
                    message=str(resp.reason),
                    headers=resp.headers,
                )
            except (
                aiohttp.ServerDisconnectedError,
                aiohttp.ClientResponseError,
            ) as ex:
                self._logger.warning(
                    "Request %d/%d: %s %r",
                    attempt_count,
                    max_attempts,
                    method,
                    _real_url(url, params),
                    exc_info=True,
                )
                last_exc = ex

            if attempt_count < max_attempts:
                should_retry, sleeping_time = self._retry_strategy(retry_delay, last_exc, resp)

                if should_retry:
                    attempt_count += 1
                    await asyncio.sleep(sleeping_time)
                    continue

            self._handle_error(last_exc, resp, resp_body)

    def _retry_strategy(
        self,
        retry_pause: float,
        ex: BaseException,
        resp: Optional[aiohttp.ClientResponse],
    ) -> Tuple[bool, float]:
        """
        Ruturn a tuple of a bool and a sleeping time.

        The bool indicate whether retry should be made.
        """

        if resp is None:
            # The issues are related to sockets or network connections.
            if isinstance(ex, aiohttp.ServerDisconnectedError):
                # retry if TCP was probably reset
                return True, retry_pause

            return False, 0

        if resp.status not in self._retry_statuses:
            return False, 0

        retry_after: Optional[int] = None
        if _RETRY_AFTER in resp.headers:
            # the remote seems busy or in maintenance
            try:
                # Support second format only
                retry_after = int(resp.headers[_RETRY_AFTER])

                if retry_after < 0:
                    retry_after = None
            except ValueError:
                # Date format is not supported
                # It usually suggests relatively long unavailability.
                return False, 0

        if retry_after is None:
            return True, retry_pause
        elif retry_after <= retry_pause:
            # below our expected limit
            return True, retry_after
        else:
            return False, 0

    def _handle_error(
        self,
        error: Exception,
        resp: Optional[aiohttp.ClientResponse],
        resp_body: bytes,
    ) -> NoReturn:
        """
        Handle errors which are not retryable.
        """

        if resp is None:
            # raise the last error if no resp available
            raise error

        try:
            message = resp_body.decode("utf-8", "strict")
        except UnicodeDecodeError:
            message = ""

        exc_class = exc._ERROR_TABLE.get(resp.status, exc.APIError)
        raise exc_class.from_response(resp, message) from error

    async def request_text(
        self,
        method: str,
        endpoint: str,
        **kwargs: Any,
    ) -> str:
        """
        Send a request and return a str.
        """

        resp = await self.request(method, endpoint, **kwargs)
        async with resp:
            return await resp.text(encoding="utf-8")

    async def request_json(
        self,
        method: str,
        endpoint: str,
        **kwargs: Any,
    ) -> Any:
        """
        Send a request and return a JSON-decoded object.
        """
        resp = await self.request(method, endpoint, **kwargs)
        async with resp:
            result = await resp.json()

        return result

    def _create_object(self, rtype: Type[T], data: Mapping[str, Any]) -> T:
        return self._mapper.create_object(rtype, data, self._context)

    def _create_list(self, rtype: Type[T], data: Sequence[Mapping[str, Any]]) -> List[T]:
        return self._mapper.create_list(rtype, data, self._context)

    def _create_dict(self, rtype: Type[T], data: Mapping[K, Mapping[str, Any]]) -> Dict[K, T]:
        return self._mapper.create_dict(rtype, data, self._context)

    @property
    def client_version(self) -> Optional[ClientVersion]:
        """qBittorrent client version"""
        return self._context.get("client_version")

    @client_version.setter
    def client_version(self, version: Optional[ClientVersion]) -> None:
        self._context["client_version"] = version

    @property
    def api_version(self) -> Optional[APIVersion]:
        """qBittorrent API version"""
        return self._context.get("api_version")

    @api_version.setter
    def api_version(self, version: Optional[APIVersion]) -> None:
        self._context["api_version"] = version

    @cached_property
    def app(self) -> "AppAPI":
        """
        Application API methods.
        """
        from aioqbt.api.app import AppAPI

        return AppAPI(self)

    @cached_property
    def auth(self) -> "AuthAPI":
        """
        Authentication API methods.
        """
        from aioqbt.api.auth import AuthAPI

        return AuthAPI(self)

    @cached_property
    def log(self) -> "LogAPI":
        """
        Log API methods.
        """
        from aioqbt.api.log import LogAPI

        return LogAPI(self)

    @cached_property
    def rss(self) -> "RSSAPI":
        """
        RSS API methods.
        """
        from aioqbt.api.rss import RSSAPI

        return RSSAPI(self)

    @cached_property
    def search(self) -> "SearchAPI":
        """
        Search API methods.
        """
        from aioqbt.api.search import SearchAPI

        return SearchAPI(self)

    @cached_property
    def sync(self) -> "SyncAPI":
        """
        Sync API methods.
        """
        from aioqbt.api.sync import SyncAPI

        return SyncAPI(self)

    @cached_property
    def torrents(self) -> "TorrentsAPI":
        """
        Torrents API methods.
        """
        from aioqbt.api.torrents import TorrentsAPI

        return TorrentsAPI(self)

    @cached_property
    def transfer(self) -> "TransferAPI":
        """
        Transfer API methods.
        """
        from aioqbt.api.transfer import TransferAPI

        return TransferAPI(self)


class APIGroup(metaclass=ABCMeta):
    """
    API group of methods.
    """

    _client_ref: Optional["APIClient"] = None

    def __init__(self, client: "APIClient"):
        # do not keep the reference if closed
        self._client_ref = None if client.is_closed() else client

    def _client(self) -> "APIClient":
        client = self._client_ref

        if client is None:
            raise RuntimeError("closed client")

        return client

    def _close(self) -> None:
        self._client_ref = None

    async def _request(
        self,
        method: str,
        endpoint: str,
        **kwargs: Any,
    ) -> aiohttp.ClientResponse:
        return await self._client().request(method, endpoint, **kwargs)

    async def _request_text(
        self,
        method: str,
        endpoint: str,
        **kwargs: Any,
    ) -> str:
        return await self._client().request_text(method, endpoint, **kwargs)

    async def _request_json(
        self,
        method: str,
        endpoint: str,
        **kwargs: Any,
    ) -> Any:
        return await self._client().request_json(method, endpoint, **kwargs)

    async def _request_mapped_object(
        self,
        rtype: Type[T],
        method: str,
        endpoint: str,
        **kwargs: Any,
    ) -> T:
        client = self._client()
        result = await client.request_json(method, endpoint, **kwargs)
        return client._create_object(rtype, result)

    async def _request_mapped_list(
        self,
        rtype: Type[T],
        method: str,
        endpoint: str,
        **kwargs: Any,
    ) -> List[T]:
        client = self._client()
        result = await client.request_json(method, endpoint, **kwargs)
        return client._create_list(rtype, result)

    async def _request_mapped_dict(
        self,
        rtype: Type[T],
        method: str,
        endpoint: str,
        **kwargs: Any,
    ) -> Dict[str, T]:
        client = self._client()
        result = await client.request_json(method, endpoint, **kwargs)
        return client._create_dict(rtype, result)


def _real_url(url: str, params: Optional[Mapping[str, Any]] = None) -> str:
    url_obj = URL(url)
    if params is not None:
        url_obj = url_obj.with_query(params)

    return str(url_obj)


async def create_client(
    url: str,
    username: Optional[str] = None,
    password: Optional[str] = None,
    *,
    logout_when_close: Optional[bool] = None,
    http: Optional[aiohttp.ClientSession] = None,
    ssl: Optional[SSLContext] = None,
) -> APIClient:
    """
    Create :class:`APIClient`.

    When both ``username`` and ``password`` are given,
    the returned client will have been successfully authenticated and automatically configured.
    Otherwise, :exc:`LoginError <aioqbt.exc.LoginError>` is raised.

    If they are omitted, :meth:`client.auth.login() <.AuthAPI.login>` need to be called manually.

    :param str url: URL to WebUI API, for example, ``https://localhost:8080/api/v2/``
    :param str username: login name
    :param str password: login password
    :param logout_when_close: whether logout during :meth:`~.APIClient.close`.
    :param http: :class:`aiohttp.ClientSession` object
    :param ssl: :class:`ssl.SSLContext` for custom TLS connections
    :raises LoginError: if authentication is failed.
    """
    if (username is None) != (password is None):
        raise TypeError("Specify both username and password arguments, or neither of them")

    if logout_when_close is None:
        logout_when_close = username is not None

    mapper = ObjectMapper()
    client = APIClient(
        base_url=url,
        mapper=mapper,
        http=http,
        ssl=ssl,
        logout_when_close=logout_when_close,
    )

    if username is not None:
        assert isinstance(password, str)
        try:
            await client.auth.login(username, password)
        except exc.LoginError:
            await client.close()
            raise

    try:
        client_version, api_version = await asyncio.gather(
            client.app.version(),
            client.app.webapi_version(),
        )
    except exc.ForbiddenError:
        pass
    else:
        client.client_version = ClientVersion.parse(client_version)
        client.api_version = APIVersion.parse(api_version)

    return client


#
# def _find_localtime():
#     from datetime import timedelta, timezone
#     from time import localtime
#
#     tm = localtime()
#     return timezone(timedelta(seconds=tm.tm_gmtoff), tm.tm_zone)
#
#
# _LOCAL_TIMEZONE = _find_localtime()


def since(version: Union[APIVersion, Tuple[int, int, int]]) -> Callable[[T], T]:
    """
    Annotate function with API version.
    """

    if not isinstance(version, APIVersion):
        version = APIVersion(*version)

    def decorator(fn: T) -> T:
        fn._api_version = version  # type: ignore[attr-defined]
        return fn

    return decorator


@overload
def virtual(fn: None) -> Callable[[T], T]:
    ...


@overload
def virtual(fn: T) -> T:
    ...


def virtual(fn: Any = None) -> Any:
    """
    Mark function not backed by endpoint.
    """
    if fn is None:
        return virtual  # pragma: no cover
    else:
        return fn



import pytest

from aioqbt.bittorrent import get_info_hash


def test_get_info_hash():
    info_hash = "0" * 40
    bin_hash = b"\x00" * 20

    assert get_info_hash(info_hash) == info_hash
    assert get_info_hash(bin_hash) == info_hash

    with pytest.raises(ValueError):
        get_info_hash("0" * 20)

    with pytest.raises(ValueError):
        get_info_hash("z" * 40)



from typing import Any, Dict, Iterable, List, Optional, Union

from typing_extensions import Literal

from aioqbt._paramdict import ParamDict
from aioqbt.api.types import (
    SearchJobResults,
    SearchJobStart,
    SearchJobStatus,
    SearchPlugin,
    SearchPluginCategory,
    SearchResultEntry,
)
from aioqbt.client import APIGroup
from aioqbt.version import APIVersion


class SearchAPI(APIGroup):
    """
    Search APIs.

    .. note::

        Search API is experimental. Methods and results may change without notice.

    """

    async def start(
        self,
        pattern: str,
        plugins: Union[List[str], Literal["all"], Literal["enabled"]],
        category: Union[str, Literal["all"]],
    ) -> SearchJobStart:
        """
        Start a search job.

        The result consists of only the search job :attr:`~.SearchJobStart.id`.

        :param pattern: Search pattern.
        :param plugins: Plugins used in search. Special values:
            ``all`` uses all plugins while ``enabled`` uses enabled plugins only.
        :param category: Search specific category or ``all``.

        :raises ~exc.ConflictError: if error.
        """
        data = ParamDict()
        data.required_str("pattern", pattern)
        data.required_list("plugins", plugins, "|")
        data.required_str("category", category)

        return await self._request_mapped_object(
            SearchJobStart,
            "POST",
            "search/start",
            data=data,
        )

    async def stop(self, id: int) -> None:
        """
        Stop a search job.

        :raises ~exc.NotFoundError: if the search job is not found.
        """
        data = ParamDict()
        data.required_int("id", id)

        await self._request_text(
            "POST",
            "search/stop",
            data=data,
        )

    async def status(self, id: Optional[int] = None) -> List[SearchJobStatus]:
        """
        Query search job statuses.

        :raises ~exc.NotFoundError: if the search job is specified but not found.
        """
        params = ParamDict()
        params.optional_int("id", id)

        return await self._request_mapped_list(
            SearchJobStatus,
            "GET",
            "search/status",
            params=params,
        )

    async def results(
        self,
        id: int,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> SearchJobResults:
        """
        Get search job results.

        :raises ~exc.NotFoundError: if the search job is not found.
        :raises ~exc.ConflictError: if ``offset`` is out of range.
        """

        params = ParamDict()
        params.required_int("id", id)
        params.optional_int("limit", limit)
        params.optional_int("offset", offset)

        client = self._client()

        data = await client.request_json(
            "GET",
            "search/results",
            params=params,
        )

        mapper = client._mapper
        context = client._context
        ret = mapper.create_object(SearchJobResults, data, context)
        ret.results = mapper.create_list(SearchResultEntry, data["results"], context)

        return ret

    async def delete(self, id: int) -> None:
        """
        Delete a search job.

        :raises ~exc.NotFoundError: if the search job is not found.
        """
        data = ParamDict()
        data.required_int("id", id)

        await self._request_text(
            "POST",
            "search/delete",
            data=data,
        )

    async def plugins(self) -> List[SearchPlugin]:
        """
        Get all plugins.
        """
        client = self._client()

        data: List[Dict[str, object]] = await client.request_json(
            "GET",
            "search/plugins",
        )

        mapper = client._mapper
        context = client._context

        result: List[SearchPlugin] = mapper.create_list(SearchPlugin, data, context)

        plugin: SearchPlugin
        plugin_data: Dict[str, Any]

        dict_list = True

        if APIVersion.compare(client.api_version, (2, 5, 2)) < 0:
            # before API 2.5.2 ~= v4.3.0alpha1
            # supportedCategories is a list of localized category name strings
            # see commit 8e8cd59d90e63b992bc5c43c29d5aec001855a4e
            for plugin in result:
                if any(isinstance(s, str) for s in plugin.supportedCategories):
                    dict_list = False
                    break

        if dict_list:
            for plugin, plugin_data in zip(result, data):
                plugin.supportedCategories = mapper.create_list(
                    SearchPluginCategory,
                    plugin_data["supportedCategories"],  # type: ignore[arg-type]
                    context,
                )

        return result

    async def install_plugin(self, sources: Iterable[str]) -> None:
        """
        Install plugins.
        """
        data = ParamDict()
        data.required_list("sources", sources, "|")

        await self._request_text(
            "POST",
            "search/installPlugin",
            data=data,
        )

    async def uninstall_plugin(self, names: Iterable[str]) -> None:
        """
        Uninstall plugins.
        """
        data = ParamDict()
        data.required_list("names", names, "|")

        await self._request_text(
            "POST",
            "search/uninstallPlugin",
            data=data,
        )

    async def enable_plugin(self, names: Iterable[str], enable: bool) -> None:
        """
        Enable/disable plugins.

        :param names: a list of plugins to enable/disable.
        :param enable: ``True`` to enable or ``False`` to disable.
        """
        data = ParamDict()
        data.required_list("names", names, "|")
        data.required_bool("enable", enable)

        await self._request_text(
            "POST",
            "search/enablePlugin",
            data=data,
        )

    async def update_plugins(self) -> None:
        """
        Update plugins.
        """
        await self._request_text(  # pragma: no cover
            "POST",
            "search/updatePlugins",
        )



import copy
import math
from datetime import timedelta
from pathlib import PurePosixPath, PureWindowsPath
from typing import Any, Union

import pytest

from aioqbt._paramdict import ParamDict
from aioqbt.chrono import TimeUnit


def test_init():
    pd = ParamDict()
    assert pd.to_dict() == {}

    data = {
        "int": 1,
        "float": 1.5,
        "str": "hello",
    }
    pd = ParamDict(data)
    expected = {
        "int": "1",
        "float": "1.5",
        "str": "hello",
    }
    assert pd.to_dict() == expected

    pd2 = ParamDict(pd)
    assert pd == pd2
    assert pd2.to_dict() == expected

    data = {
        "a": "b",
        "b": "c",
    }
    pd = ParamDict(data)
    pd["hello"] = "world"
    assert pd["hello"] == "world"

    del pd["hello"]
    with pytest.raises(KeyError):
        _ = pd["hello"]

    assert len(pd) == 2
    assert "hello" not in pd

    pd3 = copy.copy(pd)
    assert pd3 == pd


@pytest.mark.parametrize("kind", ["str", "bytes"])
def test_hash_factories(kind: str) -> None:
    def _convert_hash(value: str) -> Union[str, bytes]:
        if kind == "str":
            return value
        elif kind == "bytes":
            return bytes.fromhex(value)
        else:
            assert False, kind

    hash = "0" * 40
    hash_input = _convert_hash(hash)
    pd = ParamDict.with_hash(hash_input)
    assert pd == {
        "hash": hash,
    }

    hashes = [
        "0" * 40,
        "1" * 40,
    ]
    hashes_input = [_convert_hash(s) for s in hashes]

    pd = ParamDict.with_hashes(hashes_input)
    assert pd == {
        "hashes": "|".join(hashes),
    }

    pd = ParamDict.with_hashes_or_all(hashes_input)
    assert pd == {
        "hashes": "|".join(hashes),
    }

    pd = ParamDict.with_hashes_or_all("all")
    assert pd == {
        "hashes": "all",
    }


def test_param_names():
    pd = ParamDict()

    # test camel2snake
    with pytest.raises(TypeError, match="save_path"):
        pd.put("savePath", None)

    # test custom name
    with pytest.raises(TypeError, match="custom"):
        pd.put("savePath", None, param="custom")


def test_put():
    pd = ParamDict()
    assert len(pd) == 0

    pd.put("key", "value")
    assert pd["key"] == "value"
    assert len(pd) == 1

    pd.put("default", None, default="value2")
    assert pd["default"] == "value2"
    assert len(pd) == 2

    pd.put("optional", None, optional=True)
    assert "optional" not in pd
    assert len(pd) == 2

    with pytest.raises(TypeError):
        pd.put("required", None)

    pd.put("str", "str")
    pd.put("int", 123456)
    pd.put("float", 3.14)
    assert len(pd) == 5
    assert pd["str"] == "str"
    assert pd["int"] == "123456"
    assert pd["float"] == "3.14"

    with pytest.raises(TypeError):
        pd.put("unexpected_type", timedelta(seconds=123))

    pd.put("duration", timedelta(seconds=456), prepare=timedelta.total_seconds)

    def bad_prepare(s) -> Any:
        return None

    with pytest.raises(TypeError):
        pd.put("bad_prepare", timedelta(seconds=789), prepare=bad_prepare)


def test_put_list():
    pd = ParamDict()
    pd.optional_list("a", None, ",")
    pd.optional_list("b", ["hello", "world"], ",")
    pd.optional_list("c", ["world", "hello"], ";")

    with pytest.raises(TypeError):
        pd.required_list("missing", None, ",")  # type: ignore

    assert pd.to_dict() == {
        "b": "hello,world",
        "c": "world;hello",
    }

    with pytest.raises(ValueError):
        pd.required_list("d", [], ",", nonempty=True)


def test_put_variants():
    pd = ParamDict()
    pd.optional_str("a", None)
    pd.optional_str("b", "hello")
    pd.optional_str("c", "world")
    assert pd.to_dict() == {
        "b": "hello",
        "c": "world",
    }

    pd = ParamDict()
    pd.optional_int("a", None)
    pd.optional_int("b", 123)
    pd.required_int("c", 456)
    assert pd.to_dict() == {
        "b": "123",
        "c": "456",
    }

    pd = ParamDict()
    pd.optional_float("a", None)
    pd.optional_float("b", 123.456)
    pd.required_float("c", 456.789)
    assert pd.to_dict() == {
        "b": "123.456",
        "c": "456.789",
    }

    pd = ParamDict()
    pd.optional_duration("a", None, TimeUnit.SECONDS)
    pd.optional_duration("b", timedelta(minutes=1), TimeUnit.SECONDS)
    pd.required_duration("c", timedelta(minutes=0.5), TimeUnit.SECONDS)
    assert pd.to_dict() == {
        "b": "60",
        "c": "30",
    }

    with pytest.raises(ValueError):
        pd.required_duration("d", math.inf, TimeUnit.SECONDS)

    pd = ParamDict()
    pd.optional_path("a", None)
    pd.optional_path("b", "hello/world")
    pd.required_path("c", "hello/world")
    pd.required_path("d", PureWindowsPath("hello", "world"))
    pd.required_path("e", PurePosixPath("hello", "world"))
    pd.required_path("g", r"hello\world")

    assert pd.to_dict() == dict.fromkeys(
        ("b", "c", "d", "e", "g"),
        "hello/world",
    )



import pytest

from aioqbt.chrono import TimeUnit


def test_timeunit():
    assert TimeUnit.NANOSECONDS.from_nanos(1) == 1
    assert TimeUnit.NANOSECONDS.from_micros(1) == 1000
    assert TimeUnit.NANOSECONDS.from_millis(1) == 1_000_000
    assert TimeUnit.NANOSECONDS.from_seconds(1) == 1_000_000_000
    assert TimeUnit.SECONDS.from_seconds(1) == 1
    assert TimeUnit.SECONDS.from_minutes(1) == 60
    assert TimeUnit.SECONDS.from_hours(1) == 3600
    assert TimeUnit.SECONDS.from_days(1) == 86400

    assert TimeUnit.SECONDS.from_millis(500) == pytest.approx(0.5)
    assert TimeUnit.SECONDS.from_duration(500, TimeUnit.MILLISECONDS) == pytest.approx(0.5)

    assert TimeUnit.convert(60, TimeUnit.SECONDS, TimeUnit.MINUTES) == 1

    assert TimeUnit.MINUTES.from_duration_int(119, TimeUnit.SECONDS) == 1



from pathlib import Path
from typing import Any, List, Type, TypeVar

import pytest
from helper.lang import retry_assert
from typing_extensions import TypeGuard

from aioqbt.api.types import (
    SearchJobResults,
    SearchJobStatus,
    SearchPlugin,
    SearchPluginCategory,
    SearchResultEntry,
)
from aioqbt.client import APIClient

T = TypeVar("T")


@pytest.mark.asyncio
async def test_search(client: APIClient) -> None:
    # In this test, we install a plugin, do some search, and gather the result

    # install plugin
    # Assume the search plugin file is next to test file
    # and qBittorrent is running in the same file system.
    plugin_name = "linuxsearch"
    source = Path(__file__).with_name(f"{plugin_name}.py").as_uri()
    await client.search.install_plugin([source])

    @retry_assert
    async def plugin_installed() -> SearchPlugin:
        plugins = await client.search.plugins()

        assert isinstance(plugins, list)
        assert set([type(s) for s in plugins]) == {SearchPlugin}

        target = next((s for s in plugins if s.name == plugin_name), None)
        assert target is not None

        return target

    plugin = await plugin_installed()
    assert isinstance(plugin.enabled, bool)
    assert isinstance(plugin.fullName, str)
    assert isinstance(plugin.name, str)
    assert plugin.name == plugin_name
    assert isinstance(plugin.url, str)
    assert isinstance(plugin.version, str)
    assert isinstance(plugin.supportedCategories, list)
    assert len(plugin.supportedCategories)

    def list_type_guard(val: List[Any], tp: Type[T]) -> TypeGuard[List[T]]:
        category_types = set([type(s) for s in val])
        assert len(category_types) == 1
        return tp in category_types

    if list_type_guard(plugin.supportedCategories, SearchPluginCategory):
        # API 2.5.2 or higher
        assert "software" in set([s.id for s in plugin.supportedCategories])
    elif list_type_guard(plugin.supportedCategories, str):
        assert "software" in plugin.supportedCategories
    else:
        assert False, plugin.supportedCategories

    # start search
    search_start = await client.search.start("", [plugin_name], "all")
    assert isinstance(search_start.id, int)
    job_id = search_start.id

    # check status
    statuses = await client.search.status(job_id)
    assert isinstance(statuses, list)
    assert len(statuses) == 1
    assert isinstance(statuses[0], SearchJobStatus)
    assert isinstance(statuses[0].id, int)
    assert isinstance(statuses[0].status, str)
    assert statuses[0].status in {"Running", "Stopped"}
    assert isinstance(statuses[0].total, int)

    @retry_assert
    async def result_available():
        job = (await client.search.status(job_id))[0]
        assert job.total > 0

    await result_available()

    # check complete results
    complete = await client.search.results(job_id)
    assert isinstance(complete, SearchJobResults)
    assert isinstance(complete.status, str)
    assert complete.status in {"Running", "Stopped"}
    assert isinstance(complete.total, int)
    assert complete.total == 3
    assert isinstance(complete.results, list)
    assert len(complete.results) == complete.total
    assert set([type(s) for s in complete.results]) == {SearchResultEntry}

    # check limit results
    limited = await client.search.results(job_id, 1, 1)
    assert isinstance(limited, SearchJobResults)
    assert limited.total == complete.total
    assert len(limited.results) == 1
    assert limited.results == complete.results[1:2]

    # stop search
    await client.search.stop(job_id)

    @retry_assert
    async def job_stopped():
        statuses = await client.search.status(job_id)
        assert len(statuses) == 1
        assert statuses[0].status == "Stopped"

    await job_stopped()

    # delete search
    await client.search.delete(job_id)

    @retry_assert
    async def job_deleted():
        statuses = await client.search.status()
        assert job_id not in [s.id for s in statuses]

    await job_deleted()

    # uninstall plugin
    await client.search.uninstall_plugin([plugin_name])

    @retry_assert
    async def plugin_uninstalled():
        plugins = await client.search.plugins()
        assert plugin_name not in [s.name for s in plugins]

    await plugin_uninstalled()


@pytest.mark.asyncio
async def test_enable(client: APIClient) -> None:
    # no error response, even if plugin does not exist.
    await client.search.enable_plugin(["nonexistent"], False)


@pytest.mark.xfail(reason="Avoid internet access")
@pytest.mark.asyncio
async def test_update_plugins(client: APIClient) -> None:
    await client.search.update_plugins()



from typing import Iterable, Tuple

from aioqbt._paramdict import ParamDict
from aioqbt.api.types import TransferInfo
from aioqbt.client import APIGroup
from aioqbt.version import APIVersion

__all__ = ("TransferAPI",)


def _check_1024(name: str, num: int) -> None:
    """num must be a multiple of 1024"""
    if num % 1024 != 0:
        raise ValueError(f"{name!r} must be a multiple of 1024")


class TransferAPI(APIGroup):
    """
    API methods under ``transfer``.
    """

    async def info(self) -> TransferInfo:
        return await self._request_mapped_object(
            TransferInfo,
            "GET",
            "transfer/info",
        )

    async def speed_limits_mode(self) -> int:
        # the response is a text of either "0" or "1"
        res = await self._request_text(
            "GET",
            "transfer/speedLimitsMode",
        )
        return int(res)

    async def toggle_speed_limits_mode(self) -> None:
        await self._request_text(
            "POST",
            "transfer/toggleSpeedLimitsMode",
        )

    async def set_speed_limits_mode(self, mode: int) -> None:
        """
        Change ``speed_limits_mode``.

        If API version is less than v2.8.14, a polyfill is used to set the mode.
        """

        if APIVersion.compare(self._client().api_version, (2, 8, 14)) < 0:
            await self._set_speed_limits_mode_polyfill(mode)
            return

        # since API v2.8.14
        data = ParamDict()
        data.required_int("mode", mode)

        await self._request_text(
            "POST",
            "transfer/setSpeedLimitsMode",
            data=data,
        )

    async def _set_speed_limits_mode_polyfill(self, mode: int) -> None:
        """
        This is a polyfill that ``speed_limits_mode`` is queried and
        toggled if needed.
        """

        new_mode = bool(mode)  # a bool is also an int
        old_mode = await self.speed_limits_mode()

        if old_mode != new_mode:
            await self.toggle_speed_limits_mode()

    async def download_limit(self) -> int:
        """Get download limit (byte/second)"""
        res = await self._request_text(
            "GET",
            "transfer/downloadLimit",
        )
        return int(res)

    async def set_download_limit(self, limit: int) -> None:
        """Set download limit (byte/second)"""
        _check_1024("limit", limit)

        data = ParamDict()
        data.required_int("limit", limit)

        await self._request_text(
            "POST",
            "transfer/setDownloadLimit",
            data=data,
        )

    async def upload_limit(self) -> int:
        """Get upload limit (byte/second)"""
        res = await self._request_text(
            "GET",
            "transfer/uploadLimit",
        )
        return int(res)

    async def set_upload_limit(self, limit: int) -> None:
        """Set upload limit (byte/second)"""
        _check_1024("limit", limit)

        data = ParamDict()
        data.required_int("limit", limit)

        await self._request_text(
            "POST",
            "transfer/setUploadLimit",
            data=data,
        )

    # since API v2.3.0
    async def ban_peers(
        self,
        peers: Iterable[Tuple[str, int]],
    ) -> None:
        """
        Ban peers.

        Address may be IPv4 or IPv6 but not domain name.

        :param peers: ``(addr, port)`` pairs
        """
        pairs = []

        for host, port in peers:
            pairs.append(f"{host!s}:{port:d}")

        data = ParamDict()
        data.required_list("peers", pairs, "|")

        await self._request_text(
            "POST",
            "transfer/banPeers",
            data=data,
        )



import asyncio
from typing import Union

import aiohttp
import aiohttp.web as aiohttp_web
import pytest
from aiohttp.hdrs import RETRY_AFTER
from helper.client import LoginInfo
from helper.web import temporary_web_server

from aioqbt import exc
from aioqbt.client import APIClient, create_client


@pytest.mark.asyncio
async def test_manual_create_client(mock_login: LoginInfo):
    """Create client without logging in"""
    login = mock_login

    client = await create_client(
        url=login.url,
    )

    async with client:
        assert client.api_version is None
        await client.auth.login(login.username, login.password)


@pytest.mark.asyncio
async def test_login_failed(mock_login: LoginInfo):
    with pytest.raises(exc.LoginError):
        await create_client(
            url=mock_login.url,
            username="badUsername",
            password="badPassword",
        )


@pytest.mark.asyncio
async def test_create_client_params():
    with pytest.raises(TypeError):
        await create_client(
            "http://localhost/does/not/masster",
            username="someUser",
            password=None,
        )


@pytest.mark.asyncio
async def test_closed_errors(mock_login: LoginInfo):
    client = await create_client(
        url=mock_login.url,
        username=mock_login.username,
        password=mock_login.password,
    )
    assert not client.is_closed()

    with pytest.raises(ValueError):
        await client.request("GET", "hello", max_attempts=0)

    await client.close()
    await client.close()
    assert client.is_closed()

    with pytest.raises(RuntimeError):
        await client.request("GET", "whatever")

    with pytest.raises(RuntimeError):
        await client.torrents.info()


@pytest.mark.asyncio
async def test_server_disconnected():
    max_attempts = 3
    retry_delay = 1 / 1000  # 1 ms

    try_count = 0
    futures = set()

    async def task(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        nonlocal try_count
        try_count += 1

        # read the request header or add some delays
        # to avoid "connection reset by peer" in client side
        await reader.read(1)

        writer.close()
        await writer.wait_closed()

    def on_connected(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        futures.add(asyncio.create_task(task(reader, writer)))

    server = await asyncio.start_server(on_connected, "127.0.0.1")
    sockets = server.sockets
    host, port = sockets[0].getsockname()
    url = f"http://{host}:{port}/api/v2"

    async with APIClient(url) as client:
        with pytest.raises(aiohttp.ServerDisconnectedError):
            resp = await client.request(
                "GET",
                "hello",
                params={"value": "1"},
                retry_delay=retry_delay,
                max_attempts=max_attempts,
            )
            async with resp:
                pass

    assert try_count == max_attempts
    server.close()
    await server.wait_closed()
    await asyncio.wait(futures)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "retry_after,expected_count",
    [
        ("0", 3),
        ("9999", 1),
        ("badvalue", 1),
    ],
)
async def test_retry_after_header(retry_after: str, expected_count: int):
    max_attempts = 3
    retry_delay = 1  # 1s

    try_count = 0

    # define request handler
    async def handler(request: aiohttp_web.BaseRequest):
        nonlocal try_count
        try_count += 1

        raise aiohttp_web.HTTPServiceUnavailable(
            headers={
                RETRY_AFTER: retry_after,
            }
        )

    async with temporary_web_server(handler) as url, APIClient(url) as client:
        with pytest.raises(exc.APIError, match="Service Unavailable"):
            resp = await client.request(
                "GET",
                "hello",
                retry_delay=retry_delay,
                max_attempts=max_attempts,
            )
            async with resp:
                pass

        assert try_count == expected_count


@pytest.mark.parametrize(
    ("body", "match"),
    (
        (None, "Bad Request"),
        ("dummy dummy", "dummy dummy"),
        ("\u4e2d\u6587", "\\u4e2d\\u6587"),
        (b"\x80", "Bad Request"),
    ),
)
@pytest.mark.asyncio
async def test_error_message(body: Union[str, bytes], match):
    # define request handler
    async def handler(request: aiohttp_web.BaseRequest):
        return aiohttp_web.Response(
            status=aiohttp_web.HTTPBadRequest.status_code,
            body=body,
        )

    async with temporary_web_server(handler) as url, APIClient(url) as client:
        with pytest.raises(exc.BadRequestError, match=match):
            await client.request("GET", "/")



from typing import Any, Collection, Mapping, Set, Tuple, Type, Union

import pytest
from helper.lang import retry_assert
from helper.torrent import make_torrent_single
from helper.webapi import temporary_torrents

from aioqbt.api import AddFormBuilder
from aioqbt.api.types import SyncMainData, SyncServerState, SyncTorrentPeers
from aioqbt.client import APIClient


def value_types(value: Any) -> Union[Set[Type[Any]], Set[Tuple[Type[Any], Type[Any]]]]:
    if isinstance(value, Mapping):
        return {(type(k), type(v)) for k, v in value.items()}
    elif isinstance(value, Collection):
        return {type(v) for v in value}
    else:
        return {type(value)}


@pytest.mark.asyncio
async def test_maindata(client: APIClient) -> None:
    # Types are examined for selected keys and fail the test to update annotations

    def assert_maindata(smd: SyncMainData) -> None:
        assert isinstance(smd, SyncMainData)

        # always exist due to default/default_factory
        assert isinstance(smd.rid, int)
        assert isinstance(smd.full_update, bool)
        assert isinstance(smd.server_state, dict)
        assert_server_state(smd.server_state)

        assert isinstance(smd.torrents, dict)
        assert isinstance(smd.torrents_removed, list)
        assert value_types(smd.torrents_removed) <= {str}

        assert isinstance(smd.categories, dict)
        assert value_types(smd.categories) <= {(str, dict)}
        assert isinstance(smd.categories_removed, list)
        assert value_types(smd.categories_removed) <= {str}

        assert isinstance(smd.tags, list)
        assert value_types(smd.tags) <= {str}
        assert isinstance(smd.tags_removed, list)
        assert value_types(smd.tags_removed) <= {str}

        assert isinstance(smd.trackers, dict)
        assert value_types(smd.trackers) <= {(str, list)}
        assert isinstance(smd.trackers_removed, list)
        assert value_types(smd.trackers_removed) <= {str}

    def assert_server_state(state: SyncServerState) -> None:
        if "connection_status" in state:
            assert state["connection_status"] in {"connected", "firewalled", "disconnected"}

    maindata = await client.sync.maindata()
    assert_maindata(maindata)

    category = "sync_maindata_category"
    tag = "sync_maindata_tag"
    tracker_url = "http://localhost/tracker"

    sample = make_torrent_single(
        "sync_maindata",
        {
            "announce": tracker_url,
        },
    )

    await client.torrents.add(
        AddFormBuilder.with_client(client)
        .category(category)
        .tags([tag])
        .include_file(sample.data)
        .build()
    )

    try:

        @retry_assert
        async def ready():
            torrents = await client.torrents.info(hashes=(sample.hash,))
            assert sample.hash in [s.hash for s in torrents]

        await ready()

        maindata2 = await client.sync.maindata(maindata.rid)
        assert_maindata(maindata2)
        assert maindata2.torrents

        await client.torrents.delete((sample.hash,), delete_files=True)

        maindata3 = await client.sync.maindata(maindata2.rid)
        assert_maindata(maindata3)
        assert maindata3.torrents_removed
    finally:
        await client.torrents.delete((sample.hash,), delete_files=True)


@pytest.mark.asyncio
async def test_torrent_peers(client: APIClient):
    sample = make_torrent_single("sync_torrent_peers")

    async with temporary_torrents(client, sample):
        torrent_peers = await client.sync.torrent_peers(sample.hash)
        assert isinstance(torrent_peers, SyncTorrentPeers)
        assert isinstance(torrent_peers.rid, int)



from typing import Any

import pytest

from aioqbt.exc import MapperError
from aioqbt.mapper import ATTR_RAW_DATA, ObjectMapper, declarative, field, inspect_raw_data


@pytest.fixture
def mapper():
    return ObjectMapper()


@declarative
class Simple:
    integer: int
    string: str
    payload: Any


def test_mapper_object(mapper: ObjectMapper):
    marker = object()
    data = {
        "integer": 123,
        "string": "abc",
        "payload": "payload",
    }
    obj = mapper.create_object(Simple, data, {})

    assert obj.integer == 123
    assert obj.string == "abc"
    assert obj.payload == "payload"
    assert obj == obj, "self equality"
    assert getattr(Simple, "__hash__") is None
    assert inspect_raw_data(obj) == data

    obj2 = mapper.create_object(Simple, data, {})
    assert obj == obj2

    # extra fields
    data_extra = {
        **data,
        "extra": marker,
    }
    obj_extra = mapper.create_object(Simple, data_extra, {})

    assert obj_extra.extra == marker  # type: ignore[attr-defined]
    assert obj_extra == obj_extra
    assert obj_extra == obj
    assert inspect_raw_data(obj_extra) == data_extra

    # missing fields
    data_missing = data.copy()
    del data_missing["string"]
    obj_missing = mapper.create_object(Simple, data_missing, {})

    with pytest.raises(AttributeError):
        _ = obj_missing.string

    assert obj_missing == obj_missing
    assert obj_missing != obj
    assert obj_missing != obj_extra
    assert inspect_raw_data(obj_missing) == data_missing

    obj_missing2 = mapper.create_object(Simple, data_missing, {})
    assert obj_missing == obj_missing2


def test_mapper_list(mapper: ObjectMapper):
    data = [
        {
            "integer": s,
            "string": "hello",
        }
        for s in range(5)
    ]
    result_list = mapper.create_list(Simple, data, {})
    assert isinstance(result_list, list)
    assert len(result_list) == len(data)
    assert all(isinstance(s, Simple) for s in result_list)
    assert all(inspect_raw_data(a) == b for a, b in zip(result_list, data))
    assert all(a.integer == b["integer"] for a, b in zip(result_list, data))


def test_mapper_dict(mapper: ObjectMapper):
    marker = object()
    data = {
        "key": {
            "integer": 12345,
            "string": "hello",
            "payload": "key",
        },
        marker: {
            "integer": 12345,
            "string": "hello",
            "payload": marker,
        },
    }

    result_dict = mapper.create_dict(Simple, data, {})
    assert isinstance(result_dict, dict)
    assert result_dict.keys() == data.keys()
    for key in data.keys():
        assert isinstance(result_dict[key], Simple)
        assert inspect_raw_data(result_dict[key]) == data[key]
        assert result_dict[key].payload == key


def test_declarative_compare(mapper: ObjectMapper):
    @declarative
    class User:
        id: int
        name: str = field(compare=False)

    data = {
        "id": 0,
        "name": "root",
    }
    data2 = {
        "id": 0,
        "name": "admin",
    }
    user = mapper.create_object(User, data, {})
    user2 = mapper.create_object(User, data2, {})

    assert user == user
    assert user == user2


def test_declarative_hash(mapper: ObjectMapper):
    @declarative(unsafe_hash=True)
    class User:
        id: int
        name: str = field(hash=False)

    data = {
        "id": 0,
        "name": "root",
    }
    data2 = {
        "id": 0,
        "name": "admin",
    }
    data3 = {
        "id": 1,
        "name": "root",
    }
    user = mapper.create_object(User, data, {})
    user2 = mapper.create_object(User, data2, {})
    user3 = mapper.create_object(User, data3, {})

    assert user == user
    assert hash(user) == hash(user)

    assert user != user2
    assert hash(user) == hash(user2)

    assert user != user3
    assert hash(user) != hash(user3)


def test_declarative_repr(mapper: ObjectMapper):
    @declarative
    class Book:
        title: str
        author: str = field(repr=False)

    data = {
        "title": "Python101",
        "author": "nobody",
    }
    book = mapper.create_object(Book, data, {})

    assert "title=" in repr(book)
    assert "author=" not in repr(book)

    data2 = {
        "name": "nobody",
    }
    book2 = mapper.create_object(Book, data2, {})
    assert "title=MISSING" in repr(book2)


def test_bad_class(mapper: ObjectMapper):
    class NotDeclarative:
        pass

    with pytest.raises(ValueError):
        mapper.create_object(NotDeclarative, {}, {})


def test_bad_data(mapper: ObjectMapper):
    with pytest.raises(MapperError):
        data = {"_private": "not allowed"}
        mapper.create_object(Simple, data, {})

    with pytest.raises(LookupError):
        inspect_raw_data(object())


def test_convert_error(mapper: ObjectMapper):
    @declarative
    class Item:
        value: int = field(
            convert=lambda val, ctx: int(val),
        )

    with pytest.raises(MapperError):
        mapper.create_object(Item, {"value": "not_integer"}, {})


@declarative
class Base:
    __slots__ = ATTR_RAW_DATA


@declarative
class User(Base):
    __slots__ = ("uid", "name", "__dict__")

    uid: int
    name: str
    topic: Any


def test_slots(mapper: ObjectMapper):
    inst = mapper.create_object(Base, {}, {})
    assert isinstance(inst, Base)
    assert inspect_raw_data(inst) == {}

    data = {
        "uid": 111,
        "name": "testing",
        "topic": "python",
    }

    inst = mapper.create_object(User, data, {})
    assert isinstance(inst, User)
    assert inspect_raw_data(inst) == data
    assert inst.uid == 111
    assert inst.name == "testing"
    assert inst.topic == "python"


def test_bad_fields():
    with pytest.raises(ValueError, match="convert"):

        @declarative
        class NotCallableConvert:
            value: int = field(
                convert=object(),  # type: ignore
            )

    with pytest.raises(ValueError, match="default_factory"):

        @declarative
        class NotCallableFactory:
            value: int = field(
                default_factory=object(),  # type: ignore[call-overload]
            )

    with pytest.raises(ValueError, match="both"):

        @declarative
        class BothDefaultAndFactory:
            value: int = field(  # type: ignore[call-overload]
                default=1,
                default_factory=lambda: 2,
            )


def test_convert(mapper: ObjectMapper):
    @declarative
    class Item:
        value: int = field(
            convert=lambda val, ctx: val + 100,
        )
        value2: int = field(
            default=2,
            convert=lambda val, ctx: val + 200,
        )
        value3: int = field(
            default=3,
            convert=lambda val, ctx: val + 300,
        )

    data = {
        "value": 1,
        "value2": 2,
        "value3": 3,
    }
    inst = mapper.create_object(Item, data, {})
    assert inst.value == 101
    assert inst.value2 == 202
    assert inst.value3 == 303

    data = {
        "value": 10,
    }
    inst = mapper.create_object(Item, data, {})
    assert inst.value == 110
    assert inst.value2 == 2
    assert inst.value3 == 3



import contextlib
from typing import AsyncIterator, Dict, List

from helper.lang import busy_wait_for
from helper.torrent import TorrentData

from aioqbt.api import AddFormBuilder
from aioqbt.api.types import TorrentInfo
from aioqbt.client import APIClient


@contextlib.asynccontextmanager
async def temporary_torrents(
    client: APIClient,
    *samples: TorrentData,
    paused: bool = True,
) -> AsyncIterator[List[TorrentInfo]]:
    builder = AddFormBuilder.with_client(client)
    builder = builder.paused(paused)

    table: Dict[str, int] = {}
    for i, item in enumerate(samples):
        table[item.hash] = i
        builder = builder.include_file(item.data, f"{item.name}.torrent")

    hashes = list(table.keys())
    torrents = []

    if hashes:
        await client.torrents.add(builder.build())

        async def cond_added() -> bool:
            nonlocal torrents
            torrents = await client.torrents.info(hashes=hashes)
            return len(torrents) == len(hashes)

        success = await busy_wait_for(cond_added)
        assert success, f"Failed to add {len(samples)} torrents: {samples[0].name}"

    try:
        yield torrents
    finally:
        if hashes:
            await client.torrents.delete(hashes, True)



"""
Tools that create objects representing API results.

This module is internal to the package.
End users should not import this module directly.
"""

import builtins
import dataclasses
import sys
from dataclasses import MISSING, dataclass, is_dataclass
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import Protocol, dataclass_transform

from aioqbt.exc import MapperError

__all__ = (
    "ConvertFn",
    "ObjectMapper",
    "inspect_raw_data",
    "declarative",
    "field",
)

T = TypeVar("T")
E = TypeVar("E", bound=Enum)
K = TypeVar("K")

if TYPE_CHECKING:
    # mypy go here due to TYPE_CHECKING
    # avoid that 'type' object is not subscriptable in py37 and py38
    _FieldSequence = Sequence[dataclasses.Field[Any]]
else:
    _FieldSequence = Sequence[dataclasses.Field]

_META_CONV = "convert"

ATTR_RAW_DATA = "_aioqbt_raw_data"
ATTR_TYPE_INFO = "_aioqbt_type_info"


class _MissingRepr:
    def __repr__(self) -> str:
        """return ``MISSING`` when format/print"""
        return "MISSING"


MISSING_REPR = _MissingRepr()


class ConvertFn(Protocol):
    """
    Convert a simple value to complex one.
    """

    def __call__(self, value: Any, context: Mapping[Any, Any]) -> Any:
        pass


@dataclass
class _FieldInfo:
    convert: Optional[ConvertFn]
    default: Any
    default_factory: Optional[Callable[[], Any]]


@dataclass
class _TypeInfo(Generic[T]):
    fields: Mapping[str, _FieldInfo]
    slot_names: Sequence[str]
    default_fields: Sequence[str]


def _find_type_info(rtype: Type[T]) -> _TypeInfo[T]:
    try:
        return getattr(rtype, ATTR_TYPE_INFO)  # type: ignore[no-any-return]
    except AttributeError:
        pass

    info = _resolve_type_info(rtype)
    setattr(rtype, ATTR_TYPE_INFO, info)

    import warnings

    warnings.warn(DeprecationWarning("Use @declarative() instead"))

    return info


def _resolve_slot_names(cls: Type[T]) -> List[str]:
    # search public slot names (not prefixed with underscore)
    # modified from copyreg._slotnames()

    assert isinstance(cls, type)

    slot_names = []

    base: type
    slots: Union[str, Sequence[str]]
    for base in cls.__mro__:
        slots = getattr(base, "__slots__", ())

        if isinstance(slots, str):
            slots = (slots,)

        for name in slots:
            if name.startswith("_"):
                continue

            slot_names.append(name)

    return slot_names


def _resolve_type_info(cls: Type[T]) -> _TypeInfo[T]:
    """
    Gather dataclass information used by ObjectMapper
    """
    if not is_dataclass(cls):
        raise ValueError(f"{cls} must be a dataclass")

    slot_names = _resolve_slot_names(cls)
    fields: Dict[str, _FieldInfo] = {}
    default_fields: List[str] = []

    for field in dataclasses.fields(cls):
        f_name = field.name
        f_metadata = field.metadata

        if f_name.startswith("_"):
            continue

        conv: Optional[ConvertFn] = f_metadata.get(_META_CONV)
        if conv is not None and not callable(conv):
            raise ValueError(f"'convert' functon for field {f_name!r} must be a callable")

        default = field.default
        default_factory = None if field.default_factory is MISSING else field.default_factory

        if default_factory is not None and not callable(default_factory):
            raise ValueError(
                f"{f_name!r} expect a callable as default_factory instead of {default_factory!r}"
            )

        if default is not MISSING and default_factory is not None:
            raise ValueError(f"{f_name!r} cannot be given both default and default_factory")
        elif default is not MISSING or default_factory is not None:
            default_fields.append(f_name)

        fields[f_name] = _FieldInfo(
            convert=conv,
            default=default,
            default_factory=default_factory,
        )

    return _TypeInfo(
        fields=fields,
        slot_names=slot_names,
        default_fields=default_fields,
    )


class ObjectMapper:
    """
    Map JSON-result to Python objects.
    """

    def create_object(
        self,
        rtype: Type[T],
        data: Mapping[str, Any],
        context: Mapping[Any, Any],
    ) -> T:
        """
        Create an object from its data.
        """
        info = _find_type_info(rtype)

        dict_data = dict(data)  # copy

        # Iterate over the dict to validate key name and to convert applicable fields
        for key, value in dict_data.items():
            field_info = info.fields.get(key)

            if field_info is None:
                if key.isidentifier() and not key.startswith("_"):
                    continue
                raise MapperError(f"Bad field name: {key!r}")

            if field_info.convert is not None:
                try:
                    dict_data[key] = field_info.convert(value, context)
                except Exception as ex:
                    raise MapperError(f"Cannot convert: {key!r}={value!r}") from ex

        # Fill fields with default values
        for key in info.default_fields:
            if key in dict_data:
                continue

            field_info = info.fields[key]
            if field_info.default is not MISSING:
                dict_data[key] = field_info.default
            elif field_info.default_factory is not None:
                dict_data[key] = field_info.default_factory()
            else:
                raise AssertionError("unreachable")

        # Separate slot fields and dict fields
        slot_data = []
        for key in info.slot_names:
            if key in dict_data:
                slot_data.append((key, dict_data.pop(key)))

        # Instantiate the object
        inst = rtype.__new__(rtype)
        if dict_data:
            vars(inst).update(dict_data)

        for key, value in slot_data:
            setattr(inst, key, value)

        setattr(inst, ATTR_RAW_DATA, data)

        return inst

    def create_list(
        self,
        rtype: Type[T],
        data: Sequence[Mapping[str, Any]],
        context: Mapping[Any, Any],
    ) -> List[T]:
        """
        Create a list of objects from a list of data.
        """
        result = []

        for i, item in enumerate(data):
            result.append(self.create_object(rtype, item, context))

        return result

    def create_dict(
        self,
        rtype: Type[T],
        data: Mapping[K, Mapping[str, Any]],
        context: Mapping[Any, Any],
    ) -> Dict[K, T]:
        """
        Create a dict whose values are mapped from another dict.
        """
        result = {}

        for key, value in data.items():
            result[key] = self.create_object(rtype, value, context)

        return result


def inspect_raw_data(instance: Any) -> Dict[str, Any]:
    """Return the raw dict from which the instance is created"""
    try:
        return getattr(instance, ATTR_RAW_DATA)  # type: ignore[no-any-return]
    except AttributeError:
        raise LookupError from None


@overload
def field(
    *,
    default: T,
    init: bool = ...,
    repr: bool = ...,
    compare: bool = ...,
    hash: Optional[bool] = ...,
    convert: Optional[ConvertFn] = ...,
    metadata: Optional[Mapping[Any, Any]] = ...,
) -> T:
    ...


@overload
def field(
    *,
    default_factory: Callable[[], T],
    init: bool = ...,
    repr: bool = ...,
    compare: bool = ...,
    hash: Optional[bool] = ...,
    convert: Optional[ConvertFn] = ...,
    metadata: Optional[Mapping[Any, Any]] = ...,
) -> T:
    ...


@overload
def field(
    *,
    init: bool = ...,
    repr: bool = ...,
    compare: bool = ...,
    hash: Optional[bool] = ...,
    convert: Optional[ConvertFn] = ...,
    metadata: Optional[Mapping[Any, Any]] = ...,
) -> Any:
    ...


def field(
    *,
    default: Any = MISSING,
    default_factory: Any = MISSING,
    init: bool = True,
    repr: bool = True,
    compare: bool = True,
    hash: Optional[bool] = None,
    convert: Optional[ConvertFn] = None,
    metadata: Optional[Mapping[Any, Any]] = None,
) -> Any:
    """
    Field specifier of :func:`@declarative <.declarative>`.

    This is implemented as a wrapper of :func:`dataclasses.field` and, thus, a replacement.
    """

    # Since MISSING type is private, typing.Any is used in default and default_factory.
    # Return Any/T instead of Field[T] such that annotation matches in class definition.

    if metadata is None:
        metadata = {}
    else:
        metadata = dict(metadata)

    if convert is not None:
        if "convert" in metadata:
            raise ValueError("Cannot specify 'convert' in both argument and metadata")
        metadata["convert"] = convert

    if not metadata:
        metadata = None

    if default_factory is MISSING:
        return dataclasses.field(
            default=default,
            init=init,
            repr=repr,
            compare=compare,
            hash=hash,
            metadata=metadata,
        )
    elif default is MISSING:
        return dataclasses.field(
            default_factory=default_factory,
            init=init,
            repr=repr,
            compare=compare,
            hash=hash,
            metadata=metadata,
        )
    else:
        raise ValueError("Cannot specify both default and default_factory")


@overload
def declarative(cls: Type[T]) -> Type[T]:
    ...


@overload
def declarative(
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    unsafe_hash: bool = False,
    frozen: bool = False,
) -> Callable[[Type[T]], Type[T]]:
    ...


# PEP 681 - Data Class Transforms
# https://peps.python.org/pep-0681/
@dataclass_transform(
    field_specifiers=(
        dataclasses.Field,
        dataclasses.field,
        field,
    ),
)
def declarative(
    cls: Optional[Type[T]] = None,
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    unsafe_hash: bool = False,
    frozen: bool = False,
) -> Union[Type[T], Callable[[Type[T]], Type[T]]]:
    """
    Decorate a class to behave similarly to dataclass
    """

    def decorate(cls: Type[T]) -> Type[T]:
        cls = _process_class(
            cls,
            init=init,
            repr=repr,
            eq=eq,
            unsafe_hash=unsafe_hash,
            frozen=frozen,
        )

        type_info = _resolve_type_info(cls)
        setattr(cls, ATTR_TYPE_INFO, type_info)

        return cls

    if cls is None:
        return decorate

    return decorate(cls)


def _process_class(
    cls: Type[T],
    *,
    init: bool,
    repr: bool,
    eq: bool,
    unsafe_hash: bool,
    frozen: bool,
) -> Type[T]:
    """
    Fix AttributeError in __eq__, __hash__, __repr__ due to missing attributes.
    """

    # Note that @declarative() leverages dataclasses facilities.
    # However, ObjectMapper instantiates objects without invoking __init__().
    # If __eq__, __hash__, and __repr__ are added by dataclasses,
    # they are replaced with our ones to avoid AttributeError.

    old_methods = {s: getattr(cls, s, None) for s in ("__eq__", "__hash__", "__repr__")}

    wrap = dataclasses.dataclass(
        init=init,
        repr=repr,
        eq=eq,
        unsafe_hash=unsafe_hash,
        frozen=frozen,
    )
    cls = wrap(cls)

    def is_changed(name: str) -> bool:
        old = old_methods[name]
        new = getattr(cls, name, None)
        return new is not None and old is not new

    fields = dataclasses.fields(cls)  # type: ignore[arg-type]
    module = sys.modules.get(cls.__module__)
    globals = {} if module is None else vars(module)

    if is_changed("__eq__"):
        setattr(cls, "__eq__", _eq_fn(fields, globals=globals))

    if is_changed("__hash__"):
        setattr(cls, "__hash__", _hash_fn(fields, globals=globals))

    if is_changed("__repr__"):
        setattr(cls, "__repr__", _repr_fn(fields, globals=globals))

    return cls


def _eq_fn(
    fields: _FieldSequence,
    *,
    globals: Optional[Dict[str, object]] = None,
) -> Callable[..., str]:
    locals: Dict[str, object] = {
        "MISSING": dataclasses.MISSING,
    }
    args = ("self", "other")
    attrs = [f.name for f in fields if f.compare]
    self_tuple = _getattr_tuple_str("self", attrs, "MISSING")
    other_tuple = _getattr_tuple_str("other", attrs, "MISSING")
    body = [
        "if type(self) is type(other):",
        f" return {self_tuple} == {other_tuple}",
        "return NotImplemented",
    ]

    return _create_fn("__eq__", args, body, locals=locals, globals=globals)


def _hash_fn(
    fields: _FieldSequence,
    *,
    globals: Optional[Dict[str, object]] = None,
) -> Callable[..., int]:
    locals: Dict[str, object] = {
        "MISSING": dataclasses.MISSING,
    }
    args = ("self",)
    attrs = [f.name for f in fields if (f.compare if f.hash is None else f.hash)]
    self_tuple = _getattr_tuple_str("self", attrs, "MISSING")
    body = [f"return hash({self_tuple})"]

    return _create_fn("__hash__", args, body, locals=locals, globals=globals)


def _repr_fn(
    fields: _FieldSequence,
    *,
    globals: Optional[Dict[str, object]] = None,
) -> Callable[..., str]:
    locals: Dict[str, object] = {
        "MISSING_REPR": MISSING_REPR,
    }
    args = ("self",)
    txt = "".join(
        [f" {f.name}={{getattr(self, {f.name!r}, MISSING_REPR)!r}}" for f in fields if f.repr]
    )
    body = [
        f'return f"<{{type(self).__name__}}{txt}>"',
    ]
    fn = _create_fn("__repr__", args, body, locals=locals, globals=globals)
    from reprlib import recursive_repr

    wrap = recursive_repr()
    return wrap(fn)


# modified from dataclasses.py
def _getattr_tuple_str(
    name: str,
    attrs: Sequence[str],
    default: str,
) -> str:
    if not attrs:
        return "()"

    return f"({','.join([f'getattr({name}, {s!r}, {default})' for s in attrs])},)"


def _create_fn(
    name: str,
    args: Sequence[str],
    body: Sequence[str],
    *,
    globals: Optional[Dict[str, object]] = None,
    locals: Optional[MutableMapping[str, object]] = None,
    return_type: Any = dataclasses.MISSING,
) -> Callable[..., Any]:
    if locals is None:
        locals = {}
    if "BUILTINS" not in locals:
        locals["BUILTINS"] = builtins
    return_annotation = ""
    if return_type is not dataclasses.MISSING:
        locals["_return_type"] = return_type
        return_annotation = "->_return_type"

    args_txt = ", ".join(args)
    body_txt = "\n".join([f"  {s}" for s in body])

    # Compute the text of the entire function.
    txt = f" def {name}({args_txt}){return_annotation}:\n{body_txt}"

    local_vars = ", ".join(locals.keys())
    source = f"def _create_fn({local_vars}):\n{txt}\n return {name}"

    ns: Dict[str, Any] = {}
    exec(source, globals, ns)
    fn: Callable[..., T] = ns["_create_fn"](**locals)
    return fn



#!/usr/bin/env python3
import argparse
import contextlib
import dataclasses
import os
import re
import shlex
import subprocess
import tempfile
import textwrap
from pathlib import Path
from typing import Dict, List, Optional, Sequence

HERE = Path(__file__).parent
PROJECT_PATH = HERE.parents[1]


@dataclasses.dataclass
class Context:
    verbose: bool = False
    docker_bin: str = "docker"
    dry_run: bool = False

    tag_spec: List[str] = dataclasses.field(default_factory=list)

    # build stage
    file: Optional[str] = None
    """Dockerfile path"""

    tags: List[str] = dataclasses.field(default_factory=list)
    """Image tags"""

    image_args: Dict[str, str] = dataclasses.field(default_factory=dict)
    """--build-arg in build stage"""

    build_target: Optional[str] = None
    """--target in build stage"""

    # run stage
    run_args: List[str] = dataclasses.field(default_factory=list)
    """arguments to container in run stage"""


def get_parser() -> argparse.ArgumentParser:
    epilog = """
    Examples:

    $ python3 run-container.py --base-image debian:bookworm pytest.sh
    Run pytest.sh on debian bookworm container.

    $ python3 run.py --qbittorrent 4.6
    Run pytest.sh with qBittorrent v4.6 compiled from source.

    $ python3 run-container.py --file qbittorrent-5.0.Dockerfile \\
        --build-arg=QBITTORRENT_COMMIT=master pytest.sh
    Run pytest.sh with latest qBittorrent (assuming v5.0).
    """
    p = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent(epilog),
    )
    p.add_argument(
        "--verbose",
        "-v",
        default=False,
        action="store_true",
    )
    p.add_argument(
        "--podman",
        default=False,
        action="store_true",
        help="Use Podman instead",
    )
    p.add_argument(
        "--file",
        "-f",
        default=None,
        help="Dockerfile location (absolute or relative to this script)",
    )
    p.add_argument(
        "--tag",
        "-t",
        dest="tags",
        default=[],
        action="append",
        help="Image tags",
    )
    p.add_argument(
        "--build-target",
        default="pytest",
        help="Target in build stage",
    )
    p.add_argument(
        "--build-arg",
        dest="build_args",
        default=[],
        action="append",
        help="Image build arguments",
    )
    p.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Print commands may issue",
    )
    p.add_argument(
        "--base-image",
        help="Base image (e.g. debian:bookworm) to derive other options.",
    )
    p.add_argument(
        "--qbittorrent",
        dest="qbittorrent_version",
        help="qBittorent version (e.g. 4.6) to derive other options.",
    )
    return p


def _run(
    args: Sequence[str],
    *,
    dry_run: bool = False,
    **kwargs,
) -> int:
    assert all(isinstance(s, str) for s in args), args
    print("Command:", " ".join(shlex.quote(s) for s in args))

    if kwargs:
        print("Command kwargs:", kwargs)

    if dry_run:
        print("Command is skipped.")
        return 0

    completed = subprocess.run(
        args,
        **kwargs,
    )
    return completed.returncode


def _prepare_context(
    ns: argparse.Namespace,
    run_args: List[str],
) -> Context:
    image_args = {}
    for build_arg in ns.build_args:
        key, eq, value = build_arg.partition("=")
        if not eq:
            # Use local environment if equality sign is missing
            value = os.environ.get(key, "")
        image_args[key] = value

    run_args = list(run_args)
    if run_args and run_args[0] == "--":
        run_args.pop(0)

    ctx = Context(
        verbose=ns.verbose,
        docker_bin="podman" if ns.podman else "docker",
        file=ns.file,
        tags=ns.tags,
        build_target=ns.build_target,
        image_args=image_args,
        run_args=run_args,
    )

    if ns.qbittorrent_version is not None:
        _update_from_qbittorent(ctx, ns.qbittorrent_version)
    if ns.base_image is not None:
        _update_from_base_image(ctx, ns.base_image)

    return ctx


def _update_from_qbittorent(ctx: Context, version: str):
    """Update context from --qbittorrent"""
    match = re.match(r"\d+(?:\.\d)*", version)

    if match is None:
        raise ValueError(f"Bad version: {version}")

    if ctx.file is None:
        ctx.file = f"qbittorrent-{version}.Dockerfile"

    ctx.tag_spec.append(f"qbittorrent-{version}")


def _update_from_base_image(ctx: Context, base_image: str):
    """Update context from --base-image"""
    match = re.match(
        r"(?P<name>debian|ubuntu|pypy|python):(?P<tag>(?![.-])[a-zA-Z0-9.-]{1,128})",
        base_image,
    )

    if match is None:
        raise ValueError(f"Bad name: {base_image}")

    base_name = match.group("name")
    base_tag = match.group("tag")

    if ctx.file is None:
        ctx.file = "Dockerfile"

    fqn = f"docker.io/library/{base_name}:{base_tag}"
    ctx.image_args.setdefault("BASE_IMAGE", fqn)
    ctx.tag_spec.append(f"{base_name}-{base_tag}")


def _build_image(ctx: Context, iidfile: Path) -> str:
    """Build a container and return its image id"""
    env = {}
    if "DOCKER_BUILDKIT" not in os.environ:
        env["DOCKER_BUILDKIT"] = "1"

    args = [ctx.docker_bin, "build"]

    assert ctx.file is not None
    args.extend(("--file", ctx.file))

    if ctx.build_target is not None:
        args.extend(("--target", ctx.build_target))

    for key, value in ctx.image_args.items():
        args.extend(("--build-arg", f"{key}={value}"))

    for tag in ctx.tags:
        args.extend(("--tag", tag))

    if ctx.tag_spec:
        args.extend(("--tag", f"aioqbt/{ctx.build_target}:{'-'.join(ctx.tag_spec)}"))

    args.extend(("--iidfile", str(iidfile)))

    args.append(str(HERE))

    _run(args, check=True, env=env, cwd=str(HERE), dry_run=ctx.dry_run)

    if not ctx.dry_run:
        return iidfile.read_text()
    else:
        return "<IMAGE ID PLACEHOLDER>"


def _run_image(ctx: Context, image_id: str):
    """Run an image and return its exit code"""
    args = [
        ctx.docker_bin,
        "run",
        "-it",
        "--rm",
        f"--mount=type=bind,src={PROJECT_PATH!s},dst=/aioqbt,ro",
        "--mount=type=volume,src=aioqbt-pip-cache,dst=/root/.cache/pip",
        image_id,
        *ctx.run_args,
    ]

    return _run(args, dry_run=ctx.dry_run)


def main() -> None:
    p = get_parser()
    ns, run_args = p.parse_known_args()

    ctx = _prepare_context(ns, run_args)
    if ctx.verbose:
        print("argparse:", ns)
        print("config:", ctx)

    if ctx.file is None:
        print("No dockerfile is inferred. Try --file, --base-image, and/or --qbittorrent")
        raise SystemExit(1)

    with contextlib.ExitStack() as ctx_stack:
        if not ctx.dry_run:
            temp_prefix = "aioqbt-container-"
            temp_dir = ctx_stack.enter_context(tempfile.TemporaryDirectory(prefix=temp_prefix))
            iidfile = Path(temp_dir, "iidfile")
        else:
            iidfile = Path("/dev/null")

        image_id = _build_image(ctx, iidfile)
        print("ImageId:", image_id)

        returncode = _run_image(ctx, image_id)
        raise SystemExit(returncode)


if __name__ == "__main__":
    main()



import json
from typing import TYPE_CHECKING, Dict, List, Mapping, Optional, Union

from aioqbt._paramdict import ParamDict
from aioqbt.api.types import RSSArticle, RSSFeed, RSSFolder, RSSRule
from aioqbt.client import APIGroup

if TYPE_CHECKING:
    from aioqbt.mapper import ObjectMapper

__all__ = ("RSSAPI",)

_KEY_UID = "uid"
_KEY_URL = "url"
_KEY_ARTICLES = "articles"


def _process_item(
    mapper: "ObjectMapper",
    context: Mapping[str, object],
    data: Dict[str, object],
) -> Union[RSSFeed, RSSFolder]:
    # As we recursively walk through the data
    # tree structure is created

    uid = data.get(_KEY_UID)
    url = data.get(_KEY_URL)

    if isinstance(uid, str) and isinstance(url, str):
        feed = mapper.create_object(RSSFeed, data, context)

        articles = data.get(_KEY_ARTICLES)
        if isinstance(articles, list):
            feed.articles = mapper.create_list(RSSArticle, articles, context)
        return feed
    else:
        items = {}

        for key, value in data.items():
            assert isinstance(value, dict)
            items[key] = _process_item(mapper, context, value)

        return RSSFolder(
            _items=items,
        )


class RSSAPI(APIGroup):
    """
    RSS APIs

    .. note::

        RSS API is experimental. Methods and results may change without notice.

    """

    async def add_folder(self, path: str) -> None:
        """
        Add a new folder.

        Raise :exc:`~.exc.ConflictError` if error.
        """
        data = ParamDict()
        data.required_str("path", path)

        await self._request_text(
            "POST",
            "rss/addFolder",
            data=data,
        )

    async def add_feed(self, url: str, path: str) -> None:
        """
        Add a new feed.

        Raise :exc:`~.exc.ConflictError` if error.
        """
        data = ParamDict()
        data.required_str("url", url)
        data.required_str("path", path)

        await self._request_text(
            "POST",
            "rss/addFeed",
            data=data,
        )

    async def remove_item(self, path: str) -> None:
        """
        Add a feed/folder.

        Raise :exc:`~.exc.ConflictError` if error.
        """
        data = ParamDict()
        data.required_str("path", path)

        await self._request_text(
            "POST",
            "rss/removeItem",
            data=data,
        )

    async def move_item(self, item_path: str, dest_path: str) -> None:
        """
        Move a feed/folder.

        Raise :exc:`~.exc.ConflictError` if error.
        """
        data = ParamDict()
        data.required_str("itemPath", item_path)
        data.required_str("destPath", dest_path)

        await self._request_text(
            "POST",
            "rss/moveItem",
            data=data,
        )

    async def items(self, with_data: Optional[bool] = None) -> RSSFolder:
        """
        Get the root folder, which consists of all feeds and sub-folders.

        If ``with_data=True``, feed title and a list of articles will also be available.
        See :class:`~.RSSFeed`.
        """
        params = ParamDict()
        params.optional_bool("withData", with_data)

        client = self._client()
        result = await client.request_json(
            "GET",
            "rss/items",
            params=params,
        )

        mapper = client._mapper
        context = client._context

        root = _process_item(mapper, context, result)

        assert isinstance(root, RSSFolder)
        return root

    async def mark_as_read(self, item_path: str, article_id: Optional[str] = None) -> None:
        """
        Mark an article as read.
        """
        # API 2.5.1
        data = ParamDict()
        data.required_str("itemPath", item_path)
        data.optional_str("articleId", article_id)

        await self._request_text(
            "POST",
            "rss/markAsRead",
            data=data,
        )

    async def refresh_item(self, item_path: str) -> None:
        """
        Refresh a folder/feed.
        """
        # API 2.2.1
        data = ParamDict()
        data.required_str("itemPath", item_path)

        await self._request_text(
            "POST",
            "rss/refreshItem",
            data=data,
        )

    async def set_rule(
        self,
        rule_name: str,
        rule_def: Union[str, RSSRule, Mapping[str, object]],
    ) -> None:
        """
        Add/update a rule.
        """
        if not isinstance(rule_def, str):
            rule_def = json.dumps(dict(rule_def), separators=(",", ":"))

        data = ParamDict()
        data.required_str("ruleName", rule_name)
        data.required_str("ruleDef", rule_def)

        await self._request_text(
            "POST",
            "rss/setRule",
            data=data,
        )

    async def rename_rule(self, rule_name: str, new_rule_name: str) -> None:
        """
        Rename a rule.

        .. note::

            Before API 2.6.1, there was a bug that
            renaming rule would not change the result of :meth:`~.RSSAPI.rules`.

        """
        data = ParamDict()
        data.required_str("ruleName", rule_name)
        data.required_str("newRuleName", new_rule_name)

        await self._request_text(
            "POST",
            "rss/renameRule",
            data=data,
        )

    async def remove_rule(self, rule_name: str) -> None:
        """
        Remove a rule.
        """
        data = ParamDict()
        data.required_str("ruleName", rule_name)

        await self._request_text(
            "POST",
            "rss/removeRule",
            data=data,
        )

    async def rules(self) -> Dict[str, RSSRule]:
        """
        Get all rules.
        """
        return await self._request_json(  # type: ignore[no-any-return]
            "GET",
            "rss/rules",
        )

    async def matching_articles(self, rule_name: str) -> Dict[str, List[str]]:
        """
        Get articles matched by a rule.

        The result is a dict mapping feed names to lists of article titles.
        """
        # API 2.5.1
        params = ParamDict()
        params.required_str("ruleName", rule_name)

        return await self._request_json(  # type: ignore[no-any-return]
            "GET",
            "rss/matchingArticles",
            params=params,
        )



from typing import List, Optional

from aioqbt._paramdict import ParamDict
from aioqbt.api.types import LogMessage, LogPeer
from aioqbt.client import APIGroup

__all__ = ("LogAPI",)


class LogAPI(APIGroup):
    """
    API methods under ``log``.
    """

    async def main(
        self,
        normal: Optional[bool] = None,
        info: Optional[bool] = None,
        warning: Optional[bool] = None,
        critical: Optional[bool] = None,
        last_known_id: Optional[int] = None,
    ) -> List[LogMessage]:
        params = ParamDict()
        params.optional_bool("normal", normal)
        params.optional_bool("info", info)
        params.optional_bool("warning", warning)
        params.optional_bool("critical", critical)
        params.optional_int("last_known_id", last_known_id)

        return await self._request_mapped_list(
            LogMessage,
            "GET",
            "log/main",
            params=params,
        )

    async def peers(self, last_known_id: Optional[int] = None) -> List[LogPeer]:
        params = ParamDict()
        params.optional_int("last_known_id", last_known_id)

        return await self._request_mapped_list(
            LogPeer,
            "GET",
            "log/peers",
            params=params,
        )



import contextlib
import socket
from typing import AsyncIterator, Awaitable, Callable

from aiohttp import web

RequestHandler = Callable[[web.BaseRequest], Awaitable[web.StreamResponse]]


@contextlib.asynccontextmanager
async def temporary_site(runner: web.BaseRunner, port: int = 0) -> AsyncIterator[str]:
    # bind a TCP socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", port))

    # start server site
    site = web.SockSite(runner, sock)
    await site.start()

    # construct site URL
    host, port = sock.getsockname()
    url = f"http://localhost:{port}"

    try:
        yield url
    finally:
        # shutdown server and socket
        await site.stop()
        sock.close()


@contextlib.asynccontextmanager
async def temporary_web_server(
    handler: RequestHandler,
    port: int = 0,
) -> AsyncIterator[str]:
    # start web server
    server = web.Server(handler)
    runner = web.ServerRunner(server)
    await runner.setup()

    async with temporary_site(runner, port) as url:
        yield url



from datetime import timedelta
from io import BytesIO
from typing import Any, Callable, Dict, List, Optional, Tuple
from urllib.parse import parse_qsl

import pytest
from aiohttp import BodyPartReader, FormData, MultipartReader, hdrs
from helper.torrent import make_torrent_single

from aioqbt._compat import IntEnum, StrEnum
from aioqbt.api import AddFormBuilder


class EnumInt(IntEnum):
    MINUS_ONE = -1
    ZERO = 0
    ONE = 1


class EnumStr(StrEnum):
    EMPTY = ""
    HELLO = "hello"


class StreamWriter:
    def __init__(self, buffer: bytearray) -> None:
        self._buf = buffer

    async def write(self, buf: bytes) -> None:
        self._buf.extend(buf)


class StreamReader:
    def __init__(self, buf: bytes) -> None:
        self._buf = buf
        self._bio = BytesIO(buf)

    async def read(self, n: int = -1) -> bytes:
        return self._bio.read(n)

    async def readline(self, n: int = -1) -> bytes:
        return self._bio.readline(n)

    def unread_data(self, data: bytes) -> None:
        pos = self._bio.tell()
        start = pos - len(data)
        if start < 0:
            raise IOError("unread more than position")
        if self._buf[start:pos] != data:
            raise IOError("Data unmatched")
        self._bio.seek(start)

    def at_eof(self) -> bool:
        return self._bio.tell() == len(self._buf)


@pytest.fixture(
    scope="module",
    params=[
        "url",
        "file",
    ],
)
def builder(request: pytest.FixtureRequest) -> AddFormBuilder:
    """
    Builder may generate different form based on its payload
    """
    sample = make_torrent_single("dummy")
    param = request.param
    if param == "url":
        return AddFormBuilder().include_url(sample.magnet)
    elif param == "file":
        return AddFormBuilder().include_file(sample.data, f"{sample.name}.torrent")
    else:
        assert False, "unreachable"


async def consume_form(form: FormData) -> Dict[str, Any]:
    """
    Consume FormData and return its contents as dict
    """
    # reference: aiohttp.web_request:BaseRequest.post
    result: Dict[str, Any] = {}
    buffer = bytearray()
    writer: Any = StreamWriter(buffer)
    multipart = form()
    await multipart.write(writer)

    if multipart.content_type.startswith("multipart/"):
        stream_headers = {
            str(hdrs.CONTENT_TYPE): multipart.content_type,
        }
        stream: Any = StreamReader(buffer)
        reader = MultipartReader(stream_headers, stream)

        while True:
            part = await reader.next()

            if part is None:
                break

            if isinstance(part, BodyPartReader):
                if part.name is None:
                    raise ValueError("Missing part name?")

                data = await part.read(decode=True)

                part_ct = part.headers.get(hdrs.CONTENT_TYPE)

                if part.filename:
                    result[part.name] = (part.filename, data, part.headers)
                elif part_ct is None or part_ct.startswith("text/"):
                    charset = part.get_charset(default="utf-8")
                    result[part.name] = data.decode(charset)
                else:
                    result[part.name] = data

                await part.release()
            else:
                raise ValueError(f"Nested multipart? {type(part)}")
    else:
        charset = "utf-8"
        qsl = parse_qsl(
            buffer.decode(charset),
            keep_blank_values=True,
            encoding=charset,
        )
        result.update(qsl)

    return result


def _friendly_param_id(values: List[Tuple[str, Callable[..., Any]]]) -> List[Any]:
    result = []
    for key, meth in values:
        result.append(pytest.param(key, meth, id=meth.__name__))
    return result


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "name,method",
    [
        ("useDownloadPath", AddFormBuilder.use_download_path),
        ("skip_checking", AddFormBuilder.skip_checking),
        ("paused", AddFormBuilder.paused),
        ("root_folder", AddFormBuilder.root_folder),
        ("autoTMM", AddFormBuilder.auto_tmm),
        ("sequentialDownload", AddFormBuilder.sequential_download),
        ("firstLastPiecePrio", AddFormBuilder.first_last_piece_prio),
        ("addToTopOfQueue", AddFormBuilder.add_to_top_of_queue),
    ],
)
@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(None, None, id="None"),
        pytest.param(True, "true", id="true"),
        pytest.param(False, "false", id="false"),
    ],
)
async def test_bool_methods(
    builder: AddFormBuilder,
    name: str,
    method: Callable[[AddFormBuilder, Any], AddFormBuilder],
    value: Optional[bool],
    expected: Optional[str],
):
    data = await consume_form(method(builder, value).build())
    assert data.get(name) == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "name,method",
    [
        ("cookie", AddFormBuilder.cookie),
        ("category", AddFormBuilder.category),
        ("rename", AddFormBuilder.rename),
        ("stopCondition", AddFormBuilder.stop_condition),
        ("contentLayout", AddFormBuilder.content_layout),
    ],
)
@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(None, None, id="None"),
        pytest.param("", "", id="empty"),
        pytest.param(EnumStr.HELLO, "hello", id="StrEnum"),
    ],
)
async def test_str_methods(
    builder: AddFormBuilder,
    name: str,
    method: Callable[[AddFormBuilder, Any], AddFormBuilder],
    value: Optional[str],
    expected: Optional[str],
):
    data = await consume_form(method(builder, value).build())
    assert data.get(name) == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "key,method",
    [
        ("dlLimit", AddFormBuilder.dl_limit),
        ("upLimit", AddFormBuilder.up_limit),
    ],
)
@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(None, None, id="None"),
        pytest.param(0, "0", id="0"),
        pytest.param(-1234, "-1234", id="minus_1234"),
        pytest.param(EnumInt.ONE, "1", id="enum_one"),
    ],
)
async def test_int_methods(
    builder: AddFormBuilder,
    key: str,
    method: Callable[[AddFormBuilder, Any], AddFormBuilder],
    value: Optional[int],
    expected: Optional[str],
):
    data = await consume_form(method(builder, value).build())
    assert data.get(key) == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "key,method",
    [
        ("ratioLimit", AddFormBuilder.ratio_limit),
    ],
)
@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(None, None, id="None"),
        pytest.param(0, "0.0", id="0"),
        pytest.param(-1234, "-1234.0", id="minus_1234"),
        pytest.param(EnumInt.MINUS_ONE, "-1.0", id="enum_minus_one"),
    ],
)
async def test_float_methods(
    builder: AddFormBuilder,
    key: str,
    method: Callable[[AddFormBuilder, Any], AddFormBuilder],
    value: Optional[int],
    expected: Optional[str],
):
    data = await consume_form(method(builder, value).build())
    assert data.get(key) == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "key,method",
    [
        ("savepath", AddFormBuilder.savepath),
        ("downloadPath", AddFormBuilder.download_path),
    ],
)
@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(None, None, id="None"),
        pytest.param(r"/hello/Downloads", r"/hello/Downloads", id="slash"),
        pytest.param(r"C:\Hello\Downloads", r"C:/Hello/Downloads", id="backslash"),
    ],
)
async def test_path_methods(
    builder: AddFormBuilder,
    key: str,
    method: Callable[[AddFormBuilder, Any], AddFormBuilder],
    value: Optional[str],
    expected: Optional[str],
):
    data = await consume_form(method(builder, value).build())
    assert data.get(key) == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "key,method",
    [
        ("seedingTimeLimit", AddFormBuilder.seeding_time_limit),
        ("inactiveSeedingTimeLimit", AddFormBuilder.inactive_seeding_time_limit),
    ],
)
@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param(None, None, id="None"),
        pytest.param(0, "0", id="zero"),
        pytest.param(1234, "1234", id="int"),
        pytest.param(-1, "-1", id="negative"),
        pytest.param(timedelta(seconds=30), "0", id="timedelta_second"),
        pytest.param(timedelta(minutes=30), "30", id="timedelta_minute"),
        pytest.param(EnumInt.MINUS_ONE, "-1", id="enum_minus_one"),
    ],
)
async def test_time_limit_methods(
    builder: AddFormBuilder,
    key: str,
    method: Callable[[AddFormBuilder, Any], AddFormBuilder],
    value: Optional[int],
    expected: Optional[str],
):
    data = await consume_form(method(builder, value).build())
    assert data.get(key) == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "value,expected",
    [
        pytest.param([], "", id="0"),
        pytest.param(["hello"], "hello", id="1"),
        pytest.param(["hello", "world"], "hello,world", id="2"),
    ],
)
async def test_tags(builder: AddFormBuilder, value: Any, expected: Optional[str]):
    data = await consume_form(builder.tags(value).build())
    assert data.get("tags") == expected



"""

This module is used to detect new keys from JSON response

Rough steps:
1. Redefine some fixture to module scope
2. Use a memo object to store API type and its data
3a. For @declarative, ObjectMapperProxy will feed memo data
3b. For TypedDict, feed data manually
4. Comparing type annotations and its data received

Currently,
- New fields are reported in both @declarative types and TypedDict.
- Incompatible types are reported only in TypedDict.

"""

import asyncio
import collections
import dataclasses
import enum
import os
import warnings
from typing import (
    Any,
    AsyncIterator,
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
)

import pytest
import pytest_asyncio
from helper.client import LoginInfo, client_context
from helper.torrent import TorrentData, make_torrent_files, make_torrent_single
from helper.webapi import temporary_torrents
from typing_extensions import Literal, get_args, get_origin, get_type_hints, is_typeddict

from aioqbt import api
from aioqbt.client import APIClient
from aioqbt.mapper import ATTR_TYPE_INFO, ObjectMapper, _TypeInfo
from aioqbt.version import APIVersion

T = TypeVar("T")

_SELECTED_TYPES = [
    # app
    api.BuildInfo,
    api.Preferences,
    api.NetworkInterface,
    # torrents
    api.TorrentInfo,
    api.TorrentProperties,
    api.Tracker,
    api.WebSeed,
    api.FileEntry,
    api.Category,
    # log
    api.LogMessage,
    api.LogPeer,
    # sync
    api.SyncTorrentInfo,
    api.SyncCategory,
    api.SyncServerState,
    api.SyncMainData,
    # api.SyncPeer,
    api.SyncTorrentPeers,
    # transfer
    api.TorrentInfo,
    # RSS
    api.RSSRule,
    # api.RSSArticle,
    # api.RSSFolder,
    # api.RSSFeed,
    # Search omitted
]


class Memo:
    _data: DefaultDict[Type[Any], List[Dict[str, Any]]]

    def __init__(self) -> None:
        self._data = collections.defaultdict(list)

    def add(self, cls: Type[Any], sample: Any) -> None:
        # use Any to accept TypedDict and Dict
        assert isinstance(sample, dict), type(sample)
        self._data[cls].append(sample)

    def add_many(self, cls: Type[Any], samples: Iterable[Any]) -> None:
        # use Any to accept TypedDict and Dict
        samples = list(samples)
        assert all(isinstance(s, dict) for s in samples), {type(s) for s in samples}
        self._data[cls].extend(samples)

    def collect(self, cls: Type[Any]) -> Optional[Dict[str, List[Any]]]:
        samples = self._data.get(cls)
        if samples is None:
            return None

        result = collections.defaultdict(list)
        for entry in samples:
            for key, value in entry.items():
                result[key].append(value)

        return dict(result)


class ObjectMapperProxy(ObjectMapper):
    def __init__(self, mapper: ObjectMapper, memo: Memo) -> None:
        self.__mapper = mapper
        self.__memo = memo

    def create_object(
        self,
        rtype: Type[T],
        data: Mapping[str, Any],
        context: Mapping[Any, Any],
    ) -> T:
        self.__memo.add(rtype, dict(data))
        return self.__mapper.create_object(rtype, data, context)


@pytest.fixture(scope="module")
def warn_missing_keys() -> bool:
    try:
        return int(os.environ["WARN_MISSING_KEYS"]) == 1
    except (ValueError, KeyError, TypeError):
        return False


@pytest.fixture(scope="module")
def event_loop() -> Iterator[asyncio.AbstractEventLoop]:
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="module")
def samples() -> List[TorrentData]:
    base = {
        "announce": "http://localhost/tracker",
        "url-list": ["http://localhost/data/"],
    }
    sample_single = make_torrent_single("api_types_single", base)
    sample_files = make_torrent_files("api_types_files", base)
    return [sample_single, sample_files]


@pytest.fixture(scope="module")
def memo() -> Memo:
    return Memo()


@pytest_asyncio.fixture(scope="module")
async def client(
    mock_login: LoginInfo,
    client_cookies: List[Any],
    memo: Memo,
) -> AsyncIterator[APIClient]:
    """
    client in module scope, and store traffic to registry by proxy
    """
    client: APIClient

    async with client_context(mock_login, client_cookies) as client:
        client._mapper = ObjectMapperProxy(client._mapper, memo)
        yield client


@pytest_asyncio.fixture(scope="module")
async def activities(
    memo: Memo,
    client: APIClient,
    samples: List[TorrentData],
) -> None:
    """
    Do some activities here
    """

    hashes = [s.hash for s in samples]
    category = "test_api_types"
    rss_rule_name = "json_keys_rule"
    await client.torrents.create_category(category, "")

    async with temporary_torrents(client, *samples):
        if APIVersion.compare(client.api_version, (2, 3, 0)) >= 0:
            await client.app.build_info()
            await client.app.network_interface_list()

        await client.torrents.info(hashes=hashes)

        for hash_ in hashes:
            await client.torrents.properties(hash_)
            await client.torrents.trackers(hash_)
            await client.torrents.webseeds(hash_)
            await client.torrents.files(hash_)

        await client.torrents.categories()

        await client.transfer.info()

        await client.sync.maindata()

        for hash_ in hashes:
            await client.sync.torrent_peers(hash_)

        await client.log.main()
        await client.log.peers()

        memo.add(api.Preferences, await client.app.preferences())

        maindata = await client.sync.maindata()
        memo.add_many(api.SyncTorrentInfo, maindata.torrents.values())
        memo.add_many(api.SyncCategory, maindata.categories.values())
        memo.add(api.SyncServerState, maindata.server_state)

    await client.torrents.remove_categories([category])

    await client.rss.set_rule(rss_rule_name, api.RSSRule())
    rss_rules = await client.rss.rules()
    memo.add_many(api.RSSRule, rss_rules.values())
    await client.rss.remove_rule(rss_rule_name)


def _find_declarative_types() -> List[Type[Any]]:
    return [
        value
        for value in _SELECTED_TYPES
        if isinstance(value, type) and hasattr(value, ATTR_TYPE_INFO)
    ]


def _report_keys(cls: Type[Any], title: str, key: Iterable[str]):
    classname = cls.__name__
    msg = ", ".join(f"{s}" for s in sorted(key))
    warnings.warn(f"{title} keys in {classname}: {msg}")


@pytest.mark.asyncio
@pytest.mark.parametrize("cls", _find_declarative_types())
async def test_find_unknown_attrs(
    cls: Type[T],
    memo: Memo,
    activities: None,
    warn_missing_keys: bool,
) -> None:
    """check @declarative types"""
    collected = memo.collect(cls)
    if collected is None:
        pytest.xfail("nothing collected")

    assert dataclasses.is_dataclass(cls), cls

    try:
        type_info: _TypeInfo[T] = getattr(cls, ATTR_TYPE_INFO)
    except AttributeError:
        pytest.fail("missing type info")

    annotations = get_type_hints(cls)
    # check keys recognized by mapper
    annotations = {k: annotations[k] for k in type_info.fields.keys()}
    expected: Set[str] = set(annotations.keys())
    found: Set[str] = set(collected.keys())
    unknown = found - expected
    missing = expected - found

    # remove missing keys associated with default or default_factory
    missing.difference_update(type_info.default_fields)

    if unknown:
        _report_keys(cls, "unknown", unknown)

    if warn_missing_keys and missing:
        _report_keys(cls, "missing", missing)


def _find_typed_dict_types() -> List[Type[Any]]:
    return [value for value in _SELECTED_TYPES if is_typeddict(value)]


def _json_types(annotation: Any) -> Set[Type[Any]]:
    """
    return a set of possible Python types which JSON may be decoded into.
    Also see test_json_types() below
    """

    if isinstance(annotation, type):
        if annotation is float:
            return {float, int}
        elif annotation in (bool, int, str, list, dict, type(None)):
            return {annotation}
        elif issubclass(annotation, enum.Enum):
            if issubclass(annotation, int):
                return {int}
            elif issubclass(annotation, str):
                return {str}
            else:
                raise ValueError(f"Cannot coerce Enum {annotation} into JSON types")
        elif is_typeddict(annotation):
            # TypedDict runtime type is dict
            return {dict}

        raise ValueError(annotation)

    tp_origin = get_origin(annotation)
    tp_args = get_args(annotation)
    if isinstance(tp_origin, type):
        return _json_types(tp_origin)
    elif tp_origin == Literal:
        result = set()
        for arg in tp_args:
            result.update(_json_types(type(arg)))
        return result
    elif tp_origin == Union:
        result = set()
        for arg in tp_args:
            result.update(_json_types(arg))
        return result

    raise ValueError(annotation)


@pytest.mark.parametrize(
    "annotation,expected",
    [
        (str, {str}),
        (float, {int, float}),
        (int, {int}),
        (dict, {dict}),
        (list, {list}),
        (bool, {bool}),
        (None, {type(None)}),
        (Literal["aaa"], {str}),
        (Union[str, int], {str, int}),
        (api.TorrentState, {str}),
        (api.FilePriority, {int}),
    ],
)
def test_json_types(annotation: Any, expected: Set[Type[Any]]) -> None:
    """
    annotation := type(input)
    output := jsons.loads(json.dumps(input))
    expected := type(output)
    """

    def dummy(arg: Any) -> None:
        pass

    dummy.__annotations__["arg"] = annotation
    annotation = get_type_hints(dummy)["arg"]
    assert _json_types(annotation) == expected


@pytest.mark.asyncio
@pytest.mark.parametrize("cls", _find_typed_dict_types())
async def test_find_unknown_keys(
    cls: Type[T],
    memo: Memo,
    activities: None,
    warn_missing_keys: bool,
) -> None:
    """check TypedDict types"""

    collected = memo.collect(cls)
    if collected is None:
        pytest.xfail("nothing collected")

    assert is_typeddict(cls), cls
    annotations = get_type_hints(cls)
    expected: Set[str] = set(annotations.keys())
    found: Set[str] = set(collected.keys())
    unknown = found - expected
    missing = expected - found

    if unknown:
        _report_keys(cls, "unknown", unknown)

    if warn_missing_keys and missing:
        _report_keys(cls, "missing", missing)

    for key in expected.intersection(found):
        field_annotation = annotations[key]
        try:
            field_types = _json_types(field_annotation)
        except ValueError:
            msg = f"{cls.__name__}.{key} has unexpected annotation: {field_annotation!r}"
            pytest.fail(msg)

        sample_types = {type(s) for s in collected[key]}
        unknown_types = sample_types - field_types
        if unknown_types:
            example = next(s for s in collected[key] if type(s) in unknown_types)
            msg = (
                f"{cls.__name__}.{key} ({field_annotation}) received unknown types:"
                f" {unknown_types}; example={example!r}"
            )

            warnings.warn(msg)



import base64
import contextlib
import hashlib
import logging
import socket
import subprocess
import sys
import textwrap
import time
from pathlib import Path
from typing import Iterator, List, Optional, Sequence, Tuple, Union


def _find_free_port(socket_type: int, host: str = "") -> int:
    """Find a free port to use"""
    assert socket_type in (socket.SOCK_STREAM, socket.SOCK_DGRAM)

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind((host, 0))
    host, port = sock.getsockname()
    sock.close()

    assert isinstance(port, int)
    return port


def _wait_port_open(port: int, count: int, pause: float = 1) -> None:
    # connect to a port repeatly until it is open or maximum number of attempts is reached
    assert count >= 1, count

    for attempt in range(1, count + 1):
        error: Optional[Exception] = None

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            try:
                sock.connect(("127.0.0.1", port))
            except ConnectionRefusedError as ex:
                error = ex

        if error is None:
            return

        if attempt == count:
            raise error

        time.sleep(pause)

    raise AssertionError("unreachable")


def _combine_lists(*args: Sequence[str]) -> List[str]:
    result: List[str] = []
    for seq in args:
        assert not isinstance(seq, str), "Oops."
        result.extend(seq)

    return result


_DEFAULT_EXECUTABLE = (
    # use line-buffered stdout to capture WebUI URL and credential
    # even though it is not used currently: pytest --log-cli-level=DEBUG
    "stdbuf",
    "-oL",
    "qbittorrent-nox",
)


def _process_executable(executable: Union[str, Sequence[str], None]) -> List[str]:
    if executable is None:
        executable = _DEFAULT_EXECUTABLE

    if isinstance(executable, str):
        return [executable]
    else:
        return list(executable)


def _make_profile_conf(profile_path: Path) -> None:
    """
    Generate qBittorrent.conf in correct place

    Default user and password are set to "admin" and "adminadmin" respectively.
    """
    username = "admin"
    password = "adminadmin"

    # PBKDF-HMAC-SHA512 parameters
    salt = b"salt" * 4
    assert len(salt) == 16
    hash_name = "sha512"
    iterations = 100000

    derived_key = hashlib.pbkdf2_hmac(hash_name, password.encode("utf-8"), salt, iterations)
    dk_b64 = base64.b64encode(derived_key).decode("ascii")
    salt_b64 = base64.b64encode(salt).decode("ascii")

    txt = rf"""
    [Preferences]
    WebUI\Username="{username}"
    WebUI\Password_PBKDF2="@ByteArray({salt_b64}:{dk_b64})"
    """
    txt = textwrap.dedent(txt.lstrip("\n"))

    conf_path = profile_path.joinpath("qBittorrent/config/qBittorrent.conf")
    conf_path.parent.mkdir(exist_ok=True, parents=True)
    conf_path.write_text(txt, encoding="utf-8")


@contextlib.contextmanager
def server_process(
    profile_path: Path,
    port: Optional[int] = None,
    *,
    executable: Union[str, Sequence[str], None] = None,
    logger: Optional[logging.Logger] = None,
) -> Iterator[Tuple[str, int]]:
    if sys.platform.startswith("win"):
        raise RuntimeError("Server process is not supported on platform")

    if logger is None:
        logger = logging.getLogger("server_process")

    _make_profile_conf(profile_path)

    if port is None:
        port = _find_free_port(socket.SOCK_STREAM, "127.0.0.1")

    executable = _process_executable(executable)

    args = [
        f"--profile={profile_path}",
        f"--webui-port={port:d}",
    ]
    args = _combine_lists(executable, args)

    logger.debug("server command: %r", args)

    communication_done = False
    process = subprocess.Popen(
        args,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=0,
    )

    def log_stdout_stderr(stdout: Optional[bytes], stderr: Optional[bytes]) -> None:
        """Log stdout and stderr"""
        ts = time.time()

        stdout = stdout or b""
        stderr = stderr or b""

        assert logger is not None

        for line in stdout.splitlines():
            logger.info("server stdout %.3f: %r", ts, line)

        for line in stderr.splitlines():
            logger.error("server stderr %.3f: %r", ts, line)

    def do_communicate(input: Optional[bytes] = None, timeout: Optional[float] = None) -> None:
        """Call communicate and send outputs to logger"""

        nonlocal communication_done
        if communication_done:
            # Do not pull process after its exit.
            # it would strangely cause duplicated outputs.
            return

        try:
            stdout, stderr = process.communicate(input, timeout)
        except subprocess.TimeoutExpired as ex:
            log_stdout_stderr(ex.stdout, ex.stderr)
            raise
        else:
            # If reach here, the process has exited.
            # set the done flag
            communication_done = True
            log_stdout_stderr(stdout, stderr)

    try:
        _wait_port_open(port, 5)

        try:
            do_communicate(timeout=1 / 10)
        except subprocess.TimeoutExpired:
            pass

        # seem that the stdout is buffered on the child process.
        # WebUI URL cannot be read so "localhost" is a good guess.
        yield "localhost", port
    finally:
        logger.debug("Terminate process %d", process.pid)
        process.terminate()

        try:
            do_communicate(timeout=3)
        except subprocess.TimeoutExpired:
            logger.warning("Kill process %d", process.pid)
            process.kill()



import pytest

from aioqbt.api.types import TransferInfo
from aioqbt.client import APIClient


@pytest.mark.asyncio
async def test_speed_limits_mode(client: APIClient):
    info = await client.transfer.info()
    assert isinstance(info, TransferInfo), info
    assert isinstance(repr(info), str)

    slm = await client.transfer.speed_limits_mode()
    assert slm in (0, 1), slm

    await client.transfer.set_speed_limits_mode(slm)
    assert slm == await client.transfer.speed_limits_mode()

    await client.transfer.set_speed_limits_mode(1 - slm)
    assert slm != await client.transfer.speed_limits_mode()

    await client.transfer.set_speed_limits_mode(slm)
    assert slm == await client.transfer.speed_limits_mode()


@pytest.mark.asyncio
async def test_upload_download_limits(client: APIClient):
    up_limit = 1024 * 3
    dl_limit = 1024 * 2

    up_limit_orig = await client.transfer.upload_limit()
    await client.transfer.set_upload_limit(up_limit)
    assert up_limit == await client.transfer.upload_limit()
    await client.transfer.set_upload_limit(up_limit_orig)
    assert up_limit_orig == await client.transfer.upload_limit()

    with pytest.raises(ValueError):
        await client.transfer.set_upload_limit(1)

    dl_limit_orig = await client.transfer.download_limit()
    await client.transfer.set_download_limit(dl_limit)
    assert dl_limit == await client.transfer.download_limit()
    await client.transfer.set_download_limit(dl_limit_orig)
    assert dl_limit_orig == await client.transfer.upload_limit()

    with pytest.raises(ValueError):
        await client.transfer.set_download_limit(1)


@pytest.mark.asyncio
async def test_ban_peers(client: APIClient):
    assert client.api_version is not None
    if client.api_version < (2, 3, 0):
        pytest.skip("ban_peers requires API v2.3.0")

    await client.transfer.ban_peers((("127.0.0.1", 80),))



from datetime import datetime, timedelta, timezone

import pytest

from aioqbt.chrono import TimeUnit
from aioqbt.converter import (
    DateTimeConverter,
    DurationConverter,
    EnumConverter,
    ScalarListConverter,
)

EPOCH = datetime.fromtimestamp(0)
NONE_TABLE = dict.fromkeys(
    (
        -1,
        0xFFFF_FFFF,  # u32(-1)
    )
)


def _fromtimestamp(ts: float, tz=None):
    if ts >= 0 or tz is not None:
        return datetime.fromtimestamp(ts, tz)
    else:
        return EPOCH + timedelta(seconds=ts)


def test_datetime():
    ts = 1662803940  # full moon
    tz = timezone(timedelta(hours=8))
    tz_ctx = {
        "timezone": tz,
    }

    conv = DateTimeConverter()
    assert conv(0, {}) == _fromtimestamp(0)
    assert conv(0, tz_ctx) == _fromtimestamp(0, tz)
    assert conv(ts, {}) == _fromtimestamp(ts)
    assert conv(-ts, {}) == _fromtimestamp(-ts)
    assert conv(-1, {}) == _fromtimestamp(-1)
    assert conv(-1, tz_ctx) == _fromtimestamp(-1, tz)

    conv = DateTimeConverter(NONE_TABLE)
    assert conv(0, {}) == _fromtimestamp(0)
    assert conv(0, tz_ctx) == _fromtimestamp(0, tz)
    assert conv(ts, {}) == _fromtimestamp(ts)
    assert conv(-ts, {}) == _fromtimestamp(-ts)
    assert conv(-1, {}) is None
    assert conv(-1, tz_ctx) is None


def test_duration():
    conv = DurationConverter(TimeUnit.SECONDS)
    assert conv(0, {}) == timedelta()
    assert conv(60, {}) == timedelta(minutes=1)
    assert conv(-60, {}) == timedelta(minutes=-1)
    assert conv(-1, {}) == timedelta(seconds=-1)

    conv = DurationConverter(TimeUnit.MINUTES)
    assert conv(0, {}) == timedelta()
    assert conv(1, {}) == timedelta(minutes=1)
    assert conv(-1, {}) == timedelta(minutes=-1)

    conv = DurationConverter(TimeUnit.SECONDS, NONE_TABLE)
    assert conv(0, {}) == timedelta()
    assert conv(60, {}) == timedelta(minutes=1)
    assert conv(-60, {}) == timedelta(minutes=-1)
    assert conv(-1, {}) is None

    conv = DurationConverter(TimeUnit.MINUTES, NONE_TABLE)
    assert conv(0, {}) == timedelta()
    assert conv(1, {}) == timedelta(minutes=1)
    assert conv(-1, {}) is None


def test_scalar_list():
    conv = ScalarListConverter(",")
    assert conv("", {}) == []
    assert conv("1", {}) == ["1"]
    assert conv("1,2,3", {}) == ["1", "2", "3"]
    assert conv(",1,", {}) == ["1"]

    conv = ScalarListConverter(",", int)
    assert conv("", {}) == []
    assert conv("1", {}) == [1]
    assert conv("1,2,3", {}) == [1, 2, 3]
    assert conv(",1,", {}) == [1]

    with pytest.raises(ValueError):
        ScalarListConverter("")


def test_enums():
    from aioqbt._compat import IntEnum

    class IntChoice(IntEnum):
        ONE = 1
        TWO = 2
        THREE = 3

    conv = EnumConverter(IntChoice)
    assert conv(1, {}) is IntChoice.ONE
    with pytest.raises(ValueError):
        conv(-999, {})

    with pytest.raises(TypeError):
        EnumConverter(object)  # type: ignore



import operator
from typing import Any, Dict, List, Set, Union

BencodeTypes = Union[int, bytes, str, Dict[str, Any], List[Any]]


class BencodeError(Exception):
    pass


def dumps(obj: BencodeTypes, encoding: str = "utf-8") -> bytes:
    def iterencode(o: Any) -> Any:
        if isinstance(o, int):
            yield b"i%de" % o
        elif isinstance(o, (bytes, str)):
            if isinstance(o, str):
                o = o.encode(encoding)
            yield b"%d:" % len(o)
            yield o
        elif isinstance(o, list):
            yield b"l"
            for v in o:
                yield from iterencode(v)
            yield b"e"
        elif isinstance(o, dict):
            yield b"d"
            pairs = []
            seen: Set[bytes] = set()

            for k, v in o.items():
                if isinstance(k, str):
                    k = k.encode(encoding)
                if k in seen:
                    raise ValueError(f"Duplicated key: {k!r}")
                pairs.append((k, v))

            pairs.sort(key=operator.itemgetter(0))

            for k, v in pairs:
                yield b"%d:" % len(k)
                yield k
                yield from iterencode(v)
            yield b"e"
        else:
            raise ValueError(f"Unexpected value: {type(o)!r}")

    return b"".join(iterencode(obj))


__all__ = ("BencodeError", "dumps")



from typing import Any, AsyncIterator, Dict, Iterable, List

import pytest
import pytest_asyncio
from helper.client import LoginInfo, client_context, parse_login_env
from helper.service import server_process

from aioqbt.client import APIClient


@pytest.fixture(scope="session")
def mock_login(tmp_path_factory: pytest.TempPathFactory) -> Iterable[LoginInfo]:
    login = parse_login_env("MOCK_SERVER")
    if login is not None:
        yield login
        return

    profile_path = tmp_path_factory.mktemp("server_profile", False)
    with server_process(profile_path) as (host, port):
        yield LoginInfo(
            url=f"http://{host}:{port}/api/v2",
            username="admin",
            password="adminadmin",
        )


@pytest.fixture(scope="session")
def client_cookies() -> List[Any]:
    return []


@pytest_asyncio.fixture
async def client(mock_login: LoginInfo, client_cookies: List[Any]) -> AsyncIterator[APIClient]:
    async with client_context(mock_login, client_cookies) as client:
        yield client


@pytest_asyncio.fixture
async def temp_prefs(client: APIClient) -> AsyncIterator[Dict[str, Any]]:
    """restore preferences after change"""
    original = await client.app.preferences()

    yield dict(original)

    latest = await client.app.preferences()

    changed = {k: v for k, v in original.items() if v != latest.get(k)}

    if changed:
        await client.app.set_preferences(changed)



import contextlib
import logging
import os
from dataclasses import dataclass, field
from http.cookies import Morsel
from typing import Any, AsyncIterator, Dict, List, Optional

import aiohttp

from aioqbt.client import APIClient, create_client
from aioqbt.version import APIVersion


@dataclass
class LoginInfo:
    url: str
    username: str
    password: str
    cookies: List[Any] = field(default_factory=list)


def parse_login(spec: str) -> LoginInfo:
    from urllib.parse import urlsplit

    parts = urlsplit(spec)

    username = parts.username
    password = parts.password

    if "@" in parts.netloc:
        _, _, hostname = parts.netloc.partition("@")
    else:
        hostname = parts.netloc

    url = f"{parts.scheme}://{hostname}{parts.path}"

    assert username is not None
    assert password is not None

    return LoginInfo(
        url=url,
        username=username,
        password=password,
    )


def parse_login_env(name: str) -> Optional[LoginInfo]:
    spec = os.environ.get(name)
    if spec is None:
        return None
    return parse_login(spec)


@contextlib.asynccontextmanager
async def _cleanup_context(client: APIClient) -> AsyncIterator[None]:
    """
    Make snapshots before and after yield.
    Remove hashes, categories, and tags newly found after yield.
    """

    async def torrent_hashes() -> Dict[str, str]:
        return {s["hash"]: s["name"] for s in await client.request_json("GET", "torrents/info")}

    async def category_names() -> List[str]:
        return list((await client.request_json("GET", "torrents/categories")).keys())

    async def tag_names() -> List[str]:
        if APIVersion.compare(client.api_version, (2, 3, 0)) < 0:
            return []

        return await client.request_json("GET", "torrents/tags")  # type: ignore[no-any-return]

    hashes = await torrent_hashes()
    categories = await category_names()
    tags = await tag_names()

    yield

    hashes2 = await torrent_hashes()
    categories2 = await category_names()
    tags2 = await tag_names()

    extra_hashes = {k: v for k, v in hashes2.items() if k not in hashes}
    extra_categories = set(categories2) - set(categories)
    extra_tags = set(tags2) - set(tags)

    if extra_hashes:
        logging.warning("Clean up torrents: %r", list(extra_hashes.values()))
        await client.request_text(
            "POST",
            "torrents/delete",
            data={
                "hashes": "|".join(extra_hashes.keys()),
                "deleteFiles": "true",
            },
        )

    if extra_categories:
        logging.info("Clean up categories: %r", list(extra_categories))
        await client.request_text(
            "POST",
            "torrents/removeCategories",
            data={
                "categories": "\n".join(extra_categories),
            },
        )

    if extra_tags:
        assert APIVersion.compare(client.api_version, (2, 3, 0)) >= 0
        logging.info("Clean up tags: %r", list(extra_tags))
        await client.request_text(
            "POST",
            "torrents/deleteTags",
            data={
                "tags": ",".join(extra_tags),
            },
        )


@contextlib.asynccontextmanager
async def client_context(login: LoginInfo, cookies: List[Any]) -> AsyncIterator[APIClient]:
    """
    Save/restore cookies if any. Otherwise, log in with credential.
    """

    http = aiohttp.ClientSession(
        cookies=login.cookies,
    )

    url = login.url
    username = login.username
    password = login.password

    client: Optional[APIClient] = None

    if login.cookies:
        client = await create_client(
            url=url,
            http=http,
            logout_when_close=False,
        )

        if client.api_version is None:
            # the cookies seem expired
            client = None

    if client is None:
        client = await create_client(
            url=url,
            username=username,
            password=password,
            logout_when_close=False,
            http=http,
        )

    new_cookies = []
    for item in list(http.cookie_jar):
        assert isinstance(item, Morsel)
        new_cookies.append((item.key, item))

    cookies[:] = new_cookies

    async with http:
        async with client:
            async with _cleanup_context(client):
                yield client



import sys
from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta, timezone, tzinfo
from enum import Enum
from types import MappingProxyType
from typing import Any, Callable, Mapping, Optional, Type, TypeVar

from aioqbt._compat import StrEnum
from aioqbt.chrono import TimeUnit

__all__ = (
    "Converter",
    "NegativeMode",
    "DateTimeConverter",
    "DurationConverter",
    "EnumConverter",
    "ScalarListConverter",
)

E = TypeVar("E", bound=Enum)
_FROZEN_DICT: Mapping[Any, Any] = MappingProxyType({})


class Converter(metaclass=ABCMeta):
    """
    Convert json.decode() result to objects.
    """

    @abstractmethod
    def __call__(self, value: Any, context: Mapping[Any, Any]) -> Any:
        raise NotImplementedError


class NegativeMode(StrEnum):
    """
    Behavior when negative numbers are encountered.
    """

    ERROR = "error"  # raise
    CONVERT = "convert"  # convert negative values as well


def _find_timezone() -> Optional[tzinfo]:
    """Find local timezone"""
    # https://stackoverflow.com/a/39079819/1692260
    return datetime.now(timezone.utc).astimezone().tzinfo


_TIMEZONE = _find_timezone()
_EPOCH = datetime.fromtimestamp(0)


class DateTimeConverter(Converter):
    """
    Convert Unix timestamp to ``datetime`` object.

    .. note::
        Some attributes are roughly the same except special cases.
        - :attr:`.TorrentInfo.added_on` and :attr:`.TorrentProperties.addition_date`.
            They are set when the corresponding torrents are added and supposedly available.
        - :attr:`.TorrentInfo.seen_complete` and :attr:`.TorrentProperties.last_seen`.
            :attr:`.TorrentInfo.completion_on` and :attr:`.TorrentProperties.completion_date`.
            These attributes are directly exposed in :class:`.TorrentInfo`
            while the counterparts return invalid values as ``-1`` in
            :class:`.TorrentProperties`.

    .. note::
        Some input values are special or malformed.
        ``-1`` usually indicate the absence of value.
        In some versions, ``0xFFFF_FFFF`` may be returned because of
        the 32-bit unsigned value of ``-1``.
        However, :attr:`.TorrentProperties.creation_date` is directly
        returned regardless of its validity.
        When its value is undefined (usually represented as zero), the
        returned value usually offsets by time zone
        (e.g. ``-28800`` for UTC+8, ``28800`` for UTC-8).
    """

    # Static mappings
    _table: Mapping[int, Any]

    def __init__(self, table: Optional[Mapping[int, Any]] = None):
        super().__init__()

        if table is None:
            table = _FROZEN_DICT

        self._table = table

    def __call__(self, value: Any, context: Mapping[Any, Any]) -> Any:
        if value in self._table:
            return self._table[value]

        tz = context.get("timezone")

        if value < 0 and tz is None and sys.platform.startswith("win32"):
            # On Windows, negative timestamp without timezone may cause error
            # https://github.com/python/cpython/issues/80620#issue-1199008790
            epoch = datetime.fromtimestamp(0)
            return epoch + timedelta(seconds=value)
        else:
            return datetime.fromtimestamp(value, tz)


class RFC2822DateTimeConverter(Converter):
    """
    Convert rfc2822-formatted date time to :class:`datetime`.

    """

    def __call__(self, value: Any, context: Mapping[Any, Any]) -> datetime:
        # lazy import
        from email.utils import parsedate_to_datetime

        return parsedate_to_datetime(value)


class DurationConverter(Converter):
    """
    Convert number to timedelta.
    """

    def __init__(self, unit: TimeUnit, table: Optional[Mapping[int, Any]] = None):
        """
        :param unit: input unit
        :param table: optional special mappings
        """
        super().__init__()

        if table is None:
            table = _FROZEN_DICT

        self._unit = unit
        self._table = table

    def __call__(self, value: Any, context: Mapping[Any, Any]) -> Any:
        if value in self._table:
            return self._table[value]

        value = TimeUnit.SECONDS.from_duration(value, self._unit)
        delta = timedelta(seconds=value)
        return delta


class ScalarListConverter(Converter):
    """
    Split string into list.

    An input string is split with a separator. Empty items are discarded.
    An optional function ``cast`` may be provided to transform items.
    """

    def __init__(self, sep: str, cast: Optional[Callable[[str], Any]] = None):
        super().__init__()

        if not sep:
            raise ValueError("sep must not be empty")

        self._sep = sep
        self._cast = cast

    def __call__(self, value: Any, context: Mapping[Any, Any]) -> Any:
        items = value.split(self._sep)
        cast = self._cast

        if cast is None:
            return [s for s in items if s]
        else:
            return [cast(s) for s in items if s]


class EnumConverter(Converter):
    """
    Convert values to members of enums.
    """

    def __init__(self, enum_type: Type[E]):
        super().__init__()

        if not issubclass(enum_type, Enum):
            raise TypeError(f"expects a subclass of Enum: {enum_type!r}")

        self._enum_type = enum_type

    def __call__(self, value: Any, context: Mapping[Any, Any]) -> Any:
        try:
            return self._enum_type(value)
        except ValueError:
            raise ValueError(f"{self._enum_type} has no such value: {value!r}") from None
